
from typing import Iterator, TYPE_CHECKING

from patshared import PortMode

from .bases.elements import CanvasOptimizeIt
from .bases.group import Group
from .patchcanvas import patchcanvas

if TYPE_CHECKING:
    from .patchbay_manager import PatchbayManager


def set_group_hidden_sides(
        mng: 'PatchbayManager', group_id: int, port_mode: PortMode):
    group = mng.get_group_from_id(group_id)
    if group is None:
        return

    group.current_position.set_hidden_port_mode(
        group.current_position.hidden_port_modes() | port_mode)
    group.save_current_position()

    with CanvasOptimizeIt(mng, auto_redraw=True):
        if port_mode & PortMode.OUTPUT:
            for conn in mng.connections:
                if conn.port_out.group_id == group_id:
                    conn.remove_from_canvas()

            for portgroup in group.portgroups:
                if portgroup.port_mode is PortMode.OUTPUT:
                    portgroup.remove_from_canvas()

            for port in group.ports:
                if port.mode is PortMode.OUTPUT:
                    port.remove_from_canvas()

            for conn in mng.connections:
                if conn.port_out.group_id is group_id:
                    conn.add_to_canvas()

        if port_mode & PortMode.INPUT:
            for conn in mng.connections:
                if conn.port_in.group_id == group_id:
                    conn.remove_from_canvas()

            for portgroup in group.portgroups:
                if portgroup.port_mode is PortMode.INPUT:
                    portgroup.remove_from_canvas()

            for port in group.ports:
                if port.mode is PortMode.INPUT:
                    port.remove_from_canvas()

            for conn in mng.connections:
                if conn.port_in.group_id is group_id:
                    conn.add_to_canvas()

    mng.sg.hidden_boxes_changed.emit()

def restore_group_hidden_sides(
        mng: 'PatchbayManager', group_id: int,
        scene_pos: tuple[int, int] | None =None):
    group = mng.get_group_from_id(group_id)
    if group is None:
        return

    gpos = group.current_position
    hidden_port_mode = gpos.hidden_port_modes()
    if hidden_port_mode is PortMode.NULL:
        return

    if scene_pos is not None:
        for port_mode in PortMode.in_out_both():
            if hidden_port_mode & port_mode:
                gpos.boxes[port_mode].pos = scene_pos

    gpos.set_hidden_port_mode(PortMode.NULL)
    group.save_current_position()

    with CanvasOptimizeIt(mng):
        group.add_to_canvas()
        group.add_all_ports_to_canvas()

        for conn in mng.connections:
            if conn.port_out.group is group or conn.port_in.group is group:
                conn.add_to_canvas()

    patchcanvas.move_group_boxes(
        group.group_id, gpos,
        redraw=hidden_port_mode, restore=hidden_port_mode)
    patchcanvas.repulse_from_group(group.group_id, hidden_port_mode)

    mng.sg.hidden_boxes_changed.emit()

def restore_all_group_hidden_sides(mng: 'PatchbayManager'):
    groups_to_restore = set[Group]()

    with CanvasOptimizeIt(mng):
        for group in mng.groups:
            if group.current_position.hidden_port_modes():
                group.current_position.set_hidden_port_mode(PortMode.NULL)
                if not group.current_position.fully_set:
                    if group._is_hardware:
                        group.current_position.set_splitted(True)

                group.add_to_canvas()
                group.add_all_ports_to_canvas()
                groups_to_restore.add(group)

        # forget all hidden boxes even if these boxes are not
        # currently present in the patchbay.
        for gpos in mng.views.iter_group_poses(view_num=mng.view_number):
            gpos.set_hidden_port_mode(PortMode.NULL)

        for conn in mng.connections:
            conn.add_to_canvas()

    for group in groups_to_restore:
        patchcanvas.move_group_boxes(
            group.group_id, group.current_position,
            redraw=PortMode.BOTH, restore=PortMode.BOTH)
        patchcanvas.repulse_from_group(group.group_id, PortMode.BOTH)

    mng.sg.hidden_boxes_changed.emit()

def hide_all_groups(mng: 'PatchbayManager'):
    groups_to_hide = set[Group]()

    with CanvasOptimizeIt(mng):
        for group in mng.groups:
            if (group.current_position.hidden_port_modes()
                    is not PortMode.BOTH):
                groups_to_hide.add(group)
                group.current_position.set_hidden_port_mode(PortMode.BOTH)

        for conn in mng.connections:
            conn.remove_from_canvas()

        for group in groups_to_hide:
            for portgroup in group.portgroups:
                portgroup.remove_from_canvas()

            for port in group.ports:
                port.remove_from_canvas()

    for group in groups_to_hide:
        patchcanvas.move_group_boxes(
            group.group_id, group.current_position,
            redraw=PortMode.BOTH)

    mng.sg.hidden_boxes_changed.emit()

def list_hidden_groups(mng: 'PatchbayManager') -> Iterator[Group]:
    for group in mng.groups:
        if (group.current_position.hidden_port_modes()
                is not PortMode.NULL):
            yield group



from dataclasses import dataclass
import logging
from queue import Queue
from typing import TYPE_CHECKING, Callable

from qtpy.QtCore import QThread
from qtpy.QtGui import QGuiApplication

from patshared import PortType, PortSubType, JackMetadata

from .bases.connection import Connection
from .bases.elements import JackPortFlag, CanvasOptimizeIt
from .bases.group import Group
from .bases.port import Port
from .patchcanvas import patchcanvas

if TYPE_CHECKING:
    from .patchbay_manager import PatchbayManager


_logger = logging.getLogger(__name__)


@dataclass
class DelayedOrder:
    func: Callable
    args: tuple
    kwargs: dict
    draw_group: bool
    sort_group: bool
    clear_conns: bool
    metadata_change: bool


_delayed_orders = Queue[DelayedOrder]()


def later_by_batch(draw_group=False, sort_group=False,
                   clear_conns=False, metadata_change=False):
    '''This decorator indicates that the decorated method will be executed
    later by batch in the main thread when `_delayed_orders_timer` will
    call it.'''

    def decorator(func: Callable):
        def wrapper(*args, **kwargs):
            mng: PatchbayManager = args[0]
            if mng.very_fast_operation:
                return func(*args, **kwargs)

            _delayed_orders.put(
                DelayedOrder(func, args, kwargs,
                             draw_group or sort_group,
                             sort_group,
                             clear_conns,
                             metadata_change))

            if QThread.currentThread() is QGuiApplication.instance().thread():
                mng._delayed_orders_timer.start()
            else:
                mng.sg.out_thread_order.emit()
            return
        return wrapper
    return decorator

@later_by_batch()
def set_group_uuid_from_name(
        mng: 'PatchbayManager', client_name: str, uuid: int):
    mng.client_uuids[client_name] = uuid

    group = mng.get_group_from_name(client_name)
    if group is not None:
        group.uuid = uuid

@later_by_batch(draw_group=True)
def add_port(mng: 'PatchbayManager', name: str, port_type: PortType,
             flags: int, uuid: int) -> int:
    '''adds port and returns the group_id'''
    exst_port = mng.get_port_from_name(name)
    if exst_port is not None:
        _logger.warning(
            f'add port "{name}", '
            f'it already exists, remove it first !')

        if name in mng._ports_by_name:
            mng._ports_by_name.pop(name)
        if uuid in mng._ports_by_uuid:
            mng._ports_by_uuid.pop(uuid)

        if exst_port.type.is_jack and exst_port.uuid:
            mng.jack_metadatas.remove_uuid(exst_port.uuid)

        group = mng.get_group_from_id(exst_port.group_id)
        if group is not None:
            # remove portgroup first if port is in a portgroup
            if exst_port.portgroup_id:
                for portgroup in group.portgroups:
                    if portgroup.portgroup_id == exst_port.portgroup_id:
                        group.portgroups.remove(portgroup)
                        portgroup.remove_from_canvas()
                        break

            exst_port.remove_from_canvas()
            group.remove_port(exst_port)

    port = Port(mng, mng._next_port_id, name, port_type, flags, uuid)
    mng._next_port_id += 1

    full_port_name = name
    group_name, colon, port_name = full_port_name.partition(':')

    is_a2j_group = False
    group_is_new = False

    if (port_type is PortType.MIDI_ALSA
            and full_port_name.startswith((':ALSA_OUT:', ':ALSA_IN:'))):
        _, _alsa_key, alsa_gp_id, alsa_p_id, group_name, *rest = \
            full_port_name.split(':')
        port_name = ':'.join(rest)

        if port.flags & JackPortFlag.IS_PHYSICAL:
            is_a2j_group = True

    elif (full_port_name.startswith(('a2j:', 'Midi-Bridge:'))
            and (not mng.group_a2j_hw
                    or not port.flags & JackPortFlag.IS_PHYSICAL)):
        group_name, colon, port_name = port_name.partition(':')
        if full_port_name.startswith('a2j:'):
            if ' [' in group_name:
                group_name = group_name.rpartition(' [')[0]
            else:
                if ' (capture)' in group_name:
                    group_name = group_name.partition(' (capture)')[0]
                else:
                    group_name = group_name.partition(' (playback)')[0]

            # fix a2j wrongly substitute '.' with space
            group_name = mng.get_corrected_a2j_group_name(group_name)

        if port.flags & JackPortFlag.IS_PHYSICAL:
            is_a2j_group = True

    group = mng.get_group_from_name(group_name)
    if group is None:
        # port is in an non existing group, create the group
        gpos = mng.get_group_position(group_name)
        group = Group(mng, mng._next_group_id, group_name, gpos)

        group.a2j_group = is_a2j_group
        mng.set_group_as_nsm_client(group)

        mng._next_group_id += 1
        mng._add_group(group)

        group_is_new = True

    group.add_port(port)
    group.graceful_port(port)

    group.add_to_canvas()
    port.add_to_canvas()
    group.check_for_portgroup_on_last_port()
    group.check_for_display_name_on_last_port()

    if group_is_new:
        mng.sg.group_added.emit(group.group_id)

    return group.group_id

@later_by_batch(draw_group=True, clear_conns=True)
def remove_port(mng: 'PatchbayManager', name: str) -> int | None:
    '''remove a port from name and return its group_id'''
    port = mng.get_port_from_name(name)
    if port is None:
        return None

    if name in mng._ports_by_name:
        mng._ports_by_name.pop(name)
    if port.uuid in mng._ports_by_uuid:
        mng._ports_by_uuid.pop(port.uuid)
    if port.type.is_jack and port.uuid:
        mng.jack_metadatas.remove_uuid(port.uuid)

    group = mng.get_group_from_id(port.group_id)
    if group is None:
        return None

    # remove portgroup first if port is in a portgroup
    if port.portgroup_id:
        for portgroup in group.portgroups:
            if portgroup.portgroup_id == port.portgroup_id:
                group.portgroups.remove(portgroup)
                portgroup.remove_from_canvas()
                break

    port.remove_from_canvas()
    group.remove_port(port)

    if not group.ports:
        group.remove_from_canvas()
        mng._remove_group(group)
        mng.sg.group_removed.emit(group.group_id)
        return None

    return group.group_id

@later_by_batch(draw_group=True)
def rename_port(
        mng: 'PatchbayManager', name: str, new_name: str,
        uuid=0) -> int | None:
    if uuid:
        port = mng.get_port_from_uuid(uuid)
    else:
        port = mng.get_port_from_name(name)

    if port is None:
        if uuid:
            _logger.warning(
                f"rename_port to {new_name}, no port with uuid {uuid}")
        else:
            _logger.warning(
                f"rename_port to '{new_name}', no port named '{name}'")
        return

    # change port key in self._ports_by_name dict
    if name in mng._ports_by_name:
        mng._ports_by_name.pop(name)
    mng._ports_by_name[new_name] = port

    group_name = name.partition(':')[0]
    new_group_name = new_name.partition(':')[0]

    # In case a port rename implies another group for the port
    if group_name != new_group_name:
        group = mng.get_group_from_name(group_name)
        if group is not None:
            group.remove_port(port)
            if not group.ports:
                mng._remove_group(group)

        port.remove_from_canvas()
        port.full_name = new_name

        group = mng.get_group_from_name(new_group_name)
        if group is None:
            # copy the group_position to not move the group
            # because group has been renamed
            orig_gpos = mng.get_group_position(group_name)
            gpos = orig_gpos.copy()
            gpos.group_name = new_group_name

            group = Group(mng, mng._next_group_id, new_group_name, gpos)
            mng._next_group_id += 1
            group.add_port(port)
            group.add_to_canvas()
        else:
            group.add_port(port)

        port.add_to_canvas()
        return group.group_id

    port.full_name = new_name
    port.group.graceful_port(port)
    port.rename_in_canvas()
    return port.group.group_id

@later_by_batch(metadata_change=True)
def metadata_update(
        mng: 'PatchbayManager', uuid: int,
        key: str, value: str) -> int | None:
    '''remember metadata and return the group_id'''

    # first store metadata
    mng.jack_metadatas.add(uuid, key, value)

    if not uuid:
        # all JACK metadatas removed
        mng.pretty_diff_checker.full_update()
        mng.remove_and_add_all()
        return

    match key:
        case '':
            # all metadatas removed for an item (client or port)
            port = mng.get_port_from_uuid(uuid)
            if port is not None:
                port.rename_in_canvas()
                return port.group_id

            for group in mng.groups:
                if group.uuid == uuid:
                    return group.group_id

        case JackMetadata.ORDER:
            port = mng.get_port_from_uuid(uuid)
            if port is None:
                return

            try:
                port_order = int(value)
            except:
                _logger.warning(
                    f"JACK_METADATA_ORDER for UUID {uuid} "
                    f"value '{value}' is not an int")
                return
            else:
                port.order = port_order
                return port.group_id

        case JackMetadata.PRETTY_NAME:
            mng.pretty_diff_checker.uuid_change(uuid)
            port = mng.get_port_from_uuid(uuid)
            if port is not None:
                port.rename_in_canvas()
                return port.group_id

            for group in mng.groups:
                if group.uuid == uuid:
                    group.rename_in_canvas()
                    return group.group_id

        case JackMetadata.PORT_GROUP:
            port = mng.get_port_from_uuid(uuid)
            if port is None:
                return

            return port.group_id

        case JackMetadata.ICON_NAME:
            for group in mng.groups:
                if group.uuid == uuid:
                    group.set_client_icon(value, from_metadata=True)
                    return group.group_id

        case JackMetadata.SIGNAL_TYPE:
            port = mng.get_port_from_uuid(uuid)
            if port is None:
                return

            if port.type is PortType.AUDIO_JACK:
                if value == 'CV':
                    port.subtype = PortSubType.CV
                elif value == 'AUDIO':
                    port.subtype = PortSubType.REGULAR

            return port.group_id

@later_by_batch()
def add_connection(
        mng: 'PatchbayManager', port_out_name: str, port_in_name: str):
    port_out = mng.get_port_from_name(port_out_name)
    port_in = mng.get_port_from_name(port_in_name)

    if port_out is None or port_in is None:
        return

    for connection in mng.connections:
        if (connection.port_out is port_out
                and connection.port_in is port_in):
            return

    connection = Connection(mng, mng._next_connection_id, port_out, port_in)
    mng._next_connection_id += 1
    mng.connections.append(connection)

    connection.add_to_canvas()

@later_by_batch()
def remove_connection(
        mng: 'PatchbayManager', port_out_name: str, port_in_name: str):
    port_out = mng.get_port_from_name(port_out_name)
    port_in = mng.get_port_from_name(port_in_name)
    if port_out is None or port_in is None:
        return

    for connection in mng.connections:
        if (connection.port_out is port_out
                and connection.port_in is port_in):
            mng.connections.remove(connection)
            mng.sg.connection_removed.emit(connection.connection_id)
            connection.remove_from_canvas()
            break


def delayed_orders_timeout(mng: 'PatchbayManager'):
    '''This method is called by the QTimer self._delayed_orders_timer
    when no graph event happens during 50ms.
    It executes in the main thread all methods called since last time,
    then, it updates the canvas with new contents.'''
    _logger.debug('patchbay delayed order')

    group_ids_to_sort = set()
    some_groups_removed = False
    clear_conns = False

    with CanvasOptimizeIt(mng, auto_redraw=True):
        while _delayed_orders.qsize():
            oq = _delayed_orders.get()

            _logger.debug(f'  execute {oq.func.__name__}, {oq.args[1:]}')

            # execute the function, and get concerned group_id
            group_id = oq.func(*oq.args, **oq.kwargs)

            if oq.metadata_change and group_id is not None:
                key = oq.args[2]
                if key in (JackMetadata.ORDER, ''):
                    group_ids_to_sort.add(group_id)

            if oq.draw_group:
                if group_id is None:
                    some_groups_removed = True
            if oq.clear_conns:
                clear_conns = True

        for group in mng.groups:
            if group.group_id in group_ids_to_sort:
                group.sort_ports_in_canvas()

        if clear_conns:
            # sometimes connections are still in canvas without ports
            # probably because the message for their destruction
            # has not been received.
            # here we can assume to clear all of them
            conns_to_clean = list[Connection]()

            for conn in mng.connections:
                for port in (conn.port_out, conn.port_in):
                    fport = mng.get_port_from_name(port.full_name)
                    if fport is None:
                        conn.remove_from_canvas()
                        conns_to_clean.append(conn)
                        break

                    if not fport.in_canvas:
                        conn.remove_from_canvas()

            for conn in conns_to_clean:
                mng.connections.remove(conn)

    if some_groups_removed:
        patchcanvas.canvas.scene.resize_the_scene()

    mng.sg.patch_may_have_changed.emit()


from pathlib import Path

from qtpy.QtGui import QColor


class ThemeFile:
    path = Path()


def to_qcolor(color: str) -> QColor | None:
    ''' convert a color given with a string to a QColor.
    returns None if color has a incorrect value.'''
    if not isinstance(color, str):
        return None

    intensity_ratio = 1.0
    opacity_ratio = 1.0

    if color.startswith('-'):
        color = color.partition('-')[2].strip()
        intensity_ratio = - 1.0

    if '*' in color:
        words = color.split('*')
        next_for_opac = False

        for i, word in enumerate(words):
            if i == 0:
                color = word.strip()
                continue

            if not word:
                next_for_opac = True
                continue

            if next_for_opac:
                try:
                    opacity_ratio *= float(word.strip())
                except:
                    pass

                next_for_opac = False
                continue

            try:
                intensity_ratio *= float(word.strip())
            except:
                pass

    if color.startswith('rgb(') and color.endswith(')'):
        try:
            channels = [int(c.strip()) for c in
                        color.partition('(')[2].rpartition(')')[0].split(',')]
            assert len(channels) == 3
            qcolor = QColor(*channels)
        except:
            return None

    elif color.startswith('rgba(') and color.endswith(')'):
        try:
            values = [c.strip() for c in
                      color.partition('(')[2].rpartition(')')[0].split(',')]
            assert len(values) == 4
            qcolor = QColor(*[int(v) for v in values[:3]],
                            int(float(values[3]) * 255)) # type:ignore
        except:
            return None

    else:
        qcolor = QColor(color)

    if not qcolor.isValid():
        return None

    if intensity_ratio == 1.0 and opacity_ratio == 1.0:
        return qcolor

    if intensity_ratio < 0.0:
        qcolor = QColor(
            255 - qcolor.red(), 255 - qcolor.green(),
            255 - qcolor.blue(), qcolor.alpha())

    if opacity_ratio != 1.0:
        qcolor.setAlphaF(opacity_ratio * qcolor.alphaF())

    return qcolor.lighter(int(100 * abs(intensity_ratio)))

def rail_float(value, mini: float, maxi: float) -> float:
    return max(min(float(value), float(maxi)), float(mini))

def rail_int(value, mini: int, maxi: int) -> int:
    return max(min(int(value), int(maxi)), int(mini))




import logging
from typing import Iterator, TYPE_CHECKING

from qtpy.QtCore import QRectF
from qtpy.QtWidgets import QGraphicsItem

from patshared import (
    BoxLayoutMode, PortMode, PortType, PortSubType, BoxType)

from ..init_values import canvas, options, InlineDisplay
from ..utils import get_portgroup_name_from_ports_names

from .box_layout import PortsMinSizes, TitleOn, BoxLayout
from .box_utils import BoxStyler, TitleLine, UnwrapButton, WrappingState
from .paths_builder import build_painter_path

if TYPE_CHECKING:
    from .box_widget import (
        BoxWidget, UnwrapButton, TitleLine, WrappingState)


_logger = logging.getLogger(__name__)


def list_port_types_and_subs() -> Iterator[tuple[PortType, PortSubType]]:
    '''simple fast generator port PortType and PortSubType preventing
    incoherent couples'''
    for port_type in PortType:
        if port_type is PortType.NULL:
            continue

        for port_subtype in PortSubType:
            if ((port_subtype is PortSubType.A2J
                    and not port_type is PortType.MIDI_JACK)
                or (port_subtype is PortSubType.CV
                    and not port_type is PortType.AUDIO_JACK)):
                # No such port should exist, it is just for win some time
                continue

            yield (port_type, port_subtype)

def from_float_to(from_f: float, to_f: float, ratio: float) -> float:
    if ratio >= 1.0:
        return to_f
    if ratio <= 0.0:
        return from_f

    return from_f + (to_f - from_f) * ratio

def split_in_two(string: str, n_lines: int) -> list[str]:
    def polished_list(input_list: list[str]) -> list:
        output_list = list[str]()

        for string in input_list:
            if not string:
                continue
            if len(string) == 1:
                if output_list:
                    output_list[-1] += string
                else:
                    output_list.append(string)
            else:
                if output_list and len(output_list[-1]) <= 1:
                    output_list[-1] += string
                else:
                    output_list.append(string)

        return [s.strip() for s in output_list]

    if n_lines <= 1:
        return [string]

    sep_indexes = list[int]()
    last_was_digit = False
    last_was_upper = False

    for i, c in enumerate(string):
        if c in (' ', '-', '_', '.'):
            sep_indexes.append(i)
        else:
            if c.upper() == c:
                if c.isdigit():
                    if not last_was_digit:
                        sep_indexes.append(i)
                else:
                    if last_was_digit or not last_was_upper:
                        sep_indexes.append(i)

                last_was_upper = True
            else:
                if last_was_digit:
                    sep_indexes.append(i)
                last_was_upper = False

            last_was_digit = c.isdigit()

    if not sep_indexes:
        return [string]

    if len(sep_indexes) + 1 <= n_lines:
        return_list = list[str]()
        last_index = 0

        for sep_index in sep_indexes:
            return_list.append(string[last_index:sep_index])
            last_index = sep_index

        return_list.append(string[last_index:])

        return polished_list(return_list)

    best_indexes = [0]
    string_rest = string

    for i in range(n_lines, 1, -1):

        target = best_indexes[-1] + int(len(string_rest)/i)
        best_index = None
        best_dif = len(string)

        for s in sep_indexes:
            if s <= best_indexes[-1]:
                continue

            dif = abs(target - s)
            if dif < best_dif:
                best_index = s
                best_dif = dif
            else:
                break

        if best_index is None:
            continue

        string_rest = string[best_index:]
        best_indexes.append(best_index)

    best_indexes = best_indexes[1:]
    last_index = 0
    return_list = list[str]()

    for i in best_indexes:
        return_list.append(string[last_index:i])
        last_index = i

    return_list.append(string[last_index:])
    return polished_list(return_list)

def _get_portgroup_name(box: 'BoxWidget', portgrp_id: int):
    return get_portgroup_name_from_ports_names(
        [p.port_name for p in box._port_list if p.portgrp_id == portgrp_id])

def _should_align_port_types(box: 'BoxWidget') -> bool:
    '''check if we can align port types
    eg, align first midi input to first midi output'''
    if box.current_port_mode is not PortMode.BOTH:
        return False

    port_types_aligner = list[tuple[int, int]]()

    for port_type, port_subtype in list_port_types_and_subs():
        n_ins = 0
        n_outs = 0

        for port in box._port_list:
            if (port.port_type is port_type
                    and port.port_subtype is port_subtype):
                match port.port_mode:
                    case PortMode.INPUT:
                        n_ins += 1
                    case PortMode.OUTPUT:
                        n_outs += 1

        port_types_aligner.append((n_ins, n_outs))

    winner = PortMode.NULL

    for n_ins, n_outs in port_types_aligner:
        if ((winner is PortMode.INPUT and n_outs > n_ins)
                or (winner is PortMode.OUTPUT and n_ins > n_outs)):
            return False

        if n_ins > n_outs:
            winner = PortMode.INPUT
        elif n_outs > n_ins:
            winner = PortMode.OUTPUT

    return True

def _get_ports_min_sizes(
        box: 'BoxWidget', align_port_types: bool) -> PortsMinSizes:
    max_in_width = max_out_width = 0.0

    thm = box.get_theme()
    last_in_pos = last_out_pos = 0.0
    final_last_in_pos = final_last_out_pos = last_in_pos
    last_in_type_and_sub = (PortType.NULL, PortSubType.REGULAR)
    last_out_type_and_sub = (PortType.NULL, PortSubType.REGULAR)
    n_in_type_and_subs = 0
    n_out_type_and_subs = 0
    last_port_mode = PortMode.NULL

    for port_type, port_subtype in list_port_types_and_subs():
        for port in box._port_list:
            if (port.port_type is not port_type
                    or port.port_subtype is not port_subtype):
                continue

            last_of_portgrp = bool(port.pg_pos + 1 == port.pg_len)
            size = 0
            max_pwidth = options.max_port_width

            if port.port_mode is PortMode.INPUT:
                port_offset = thm.port_in_offset
            else:
                port_offset = thm.port_out_offset

            if port.portgrp_id:
                portgrp = canvas.get_portgroup(box._group_id, port.portgrp_id)
                if port.pg_pos == 0:
                    portgrp_name = _get_portgroup_name(box, port.portgrp_id)

                    if portgrp is not None and portgrp.widget is not None:
                        portgrp.widget.set_print_name(
                            portgrp_name,
                            max_pwidth - canvas.theme.port_grouped_width - 5)

                port.widget.set_print_name(
                    port.port_name.replace(
                        _get_portgroup_name(box, port.portgrp_id), '', 1),
                    int(max_pwidth / 2))

                if portgrp is None or portgrp.widget is None:
                    _logger.warning('_get_ports_min_sizes, '
                                    'no portgrp or no portgrp.widget')
                    continue

                if (portgrp.widget.get_text_width() + 5
                        > max_pwidth - port.widget.get_text_width()):
                    portgrp.widget.reduce_print_name(
                        max_pwidth - int(port.widget.get_text_width()) - 5)

                # the port_grouped_width is also used to define
                # the portgroup minimum width
                size = (max(portgrp.widget.get_text_width() + 6.0,
                            canvas.theme.port_grouped_width)
                        + max(port.widget.get_text_width() + 6.0,
                                canvas.theme.port_grouped_width)
                        + port_offset)
            else:
                port.widget.set_print_name(port.port_name, max_pwidth)
                size = max(port.widget.get_text_width() + 6.0 + port_offset, 20.0)

            type_and_sub = (port.port_type, port.port_subtype)

            if port.port_mode is PortMode.INPUT:
                max_in_width = max(max_in_width, size)
                if type_and_sub != last_in_type_and_sub:
                    if (last_in_type_and_sub
                            != (PortType.NULL, PortSubType.REGULAR)):
                        last_in_pos += thm.port_type_spacing
                    last_in_type_and_sub = type_and_sub
                    n_in_type_and_subs += 1

                last_in_pos += canvas.theme.port_height
                if last_of_portgrp:
                    last_in_pos += thm.port_spacing

            elif port.port_mode is PortMode.OUTPUT:
                max_out_width = max(max_out_width, size)

                if type_and_sub != last_out_type_and_sub:
                    if (last_out_type_and_sub !=
                            (PortType.NULL, PortSubType.REGULAR)):
                        last_out_pos += thm.port_type_spacing
                    last_out_type_and_sub = type_and_sub
                    n_out_type_and_subs += 1

                last_out_pos += canvas.theme.port_height
                if last_of_portgrp:
                    last_out_pos += thm.port_spacing

            final_last_in_pos = last_in_pos
            final_last_out_pos = last_out_pos

        if align_port_types:
            # align port types horizontally
            if last_in_pos > last_out_pos:
                last_out_type_and_sub = last_in_type_and_sub
            else:
                last_in_type_and_sub = last_out_type_and_sub
            last_in_pos = last_out_pos = max(last_in_pos, last_out_pos)

    # calculates height in case of one column only
    last_inout_pos = 0.0
    last_type_and_sub = (PortType.NULL, PortSubType.REGULAR)
    n_inout_types_and_sub = 0

    if box.current_port_mode is PortMode.BOTH:
        for port_type, port_subtype in list_port_types_and_subs():
            for port in box._port_list:
                if (port.port_type is not port_type
                        or port.port_subtype is not port_subtype):
                    continue

                if (port.port_type, port.port_subtype) != last_type_and_sub:
                    if last_type_and_sub != (PortType.NULL, PortSubType.REGULAR):
                        last_inout_pos += thm.port_type_spacing
                    last_type_and_sub = (port.port_type, port.port_subtype)
                    n_inout_types_and_sub += 1

                if port.pg_pos:
                    continue

                last_inout_pos += port.pg_len * canvas.theme.port_height
                last_inout_pos += thm.port_spacing

                last_port_mode = port.port_mode

    return PortsMinSizes(
        final_last_in_pos,
        final_last_out_pos,
        last_inout_pos,
        max_in_width + canvas.theme.port_height / 2.0,
        max_out_width + canvas.theme.port_height / 2.0,
        n_in_type_and_subs,
        n_out_type_and_subs,
        n_inout_types_and_sub,
        last_port_mode
    )

def _split_title(box: 'BoxWidget', n_lines: int) -> tuple[TitleLine, ...]:
    title, slash, subtitle = box._group_name.partition('/')

    if (not subtitle
            and box._box_type == BoxType.CLIENT
            and ' (' in box._group_name
            and box._group_name.endswith(')')):
        title, parenthese, subtitle = box._group_name.partition(' (')
        subtitle = subtitle[:-1]

    theme = box.get_theme()

    if box._box_type == BoxType.CLIENT and subtitle:
        # if there is a subtitle, title is not bold when subtitle is.
        # so title is 'little'
        client_line = TitleLine(title, theme, little=True)
        subclient_line = TitleLine(subtitle, theme)
        title_lines = []

        if n_lines <= 2:
            title_lines.append(client_line)
            title_lines.append(subclient_line)

        else:
            if client_line.size > subclient_line.size:
                client_strs = split_in_two(title, 2)
                for client_str in client_strs:
                    title_lines.append(
                        TitleLine(client_str, theme, little=True))

                for subclient_str in split_in_two(subtitle, n_lines - 2):
                    title_lines.append(TitleLine(subclient_str, theme))
            else:
                two_lines_title = False

                if n_lines >= 4:
                    # Check if we need to split the client title
                    # it could be "Carla-Multi-Client.Carla".
                    subtitles = split_in_two(subtitle, n_lines - 2)

                    for subtt in subtitles:
                        subtt_line = TitleLine(subtt, theme)
                        if subtt_line.size > client_line.size:
                            break
                    else:
                        client_strs = split_in_two(title, 2)
                        for client_str in client_strs:
                            title_lines.append(
                                TitleLine(client_str, theme, little=True))
                        two_lines_title = True

                if not two_lines_title:
                    title_lines.append(client_line)

                subt_len = n_lines - 1
                if two_lines_title:
                    subt_len -= 1
                    titles = split_in_two(subtitle, subt_len)
                    for title in titles:
                        title_lines.append(TitleLine(title, theme))
                else:
                    titles = split_in_two('uuuu' + subtitle, subt_len)

                    # supress the uuuu
                    for i, title in enumerate(titles):
                        if i == 0:
                            title = title[4:]
                            if not title:
                                continue
                        title_lines.append(TitleLine(title, theme))
    else:
        if n_lines >= 2:
            titles = list[str]()
            if (box._group_name.startswith(
                    ('Carla.', 'Carla-Multi-Client.', 'Carla-Single-Client.'))
                    and '/' in box._group_name):
                first_line, slash, last_line = box._group_name.partition('/')
                titles = [first_line + '/'] + \
                    split_in_two(last_line, n_lines - 1)
            else:
                titles = split_in_two(box._group_name, n_lines)

            title_lines = [TitleLine(tt, theme) for tt in titles]
        else:
            title_lines = [TitleLine(box._group_name, theme)]

    return tuple(title_lines)

def _choose_box_layout(
        box: 'BoxWidget',
        ports_min_sizes: PortsMinSizes) -> tuple[BoxLayout, BoxLayout]:
    '''choose in how many lines the title should be splitted
    and if the box layout should be large or high.
    Return the 2 best layout choices.'''

    box_theme = box.get_theme()
    font_size = box_theme.font.pixelSize()

    if box.has_top_icon:
        icon_size = int(box_theme.icon_size)
    else:
        icon_size = 0

    # Check Text Name size

    # first look in cache if title sizes are stocked
    all_title_templates = box_theme.get_title_templates(
        box._group_name, icon_size)
    lines_choice_max = len(all_title_templates) - 1

    if not all_title_templates:
        # this box title is not in cache, we need to estimate all its sizes
        # depending number of splitted lines.
        title_template = {
            "title_width": 0,
            "header_width": 0,
            "title_height": 0,
            "header_height": 0}

        all_title_templates = [title_template.copy() for i in range(8)]

        last_lines_count = 0
        title_line_y_start = 1 + font_size
        title_height = title_line_y_start + 3

        if box.has_top_icon:
            header_height = 3 + icon_size + 3
        else:
            header_height = title_height

        i = 0
        for i in range(1, 8):
            max_title_width = 0
            max_header_width = 50
            if box._plugin_inline is not InlineDisplay.DISABLED:
                max_header_width = 200

            title_lines = _split_title(box, i)

            for j, title_line in enumerate(title_lines):
                title_line.y = title_line_y_start + j * int(font_size * 1.4)
                max_title_width = int(max(max_title_width, title_line.size))
                header_width = title_line.size + 10

                if (box.has_top_icon
                        and title_line.y <= icon_size + 6 + font_size):
                    # text line is at right of the icon
                    header_width += icon_size + 4

                max_header_width = max(max_header_width, int(header_width))

            title_height = int(
                title_line_y_start
                + int(font_size * 1.4) * (len(title_lines) - 1)
                + font_size / 2)

            header_height = max(header_height, title_height)

            new_title_template = title_template.copy()
            new_title_template['title_width'] = max_title_width
            new_title_template['header_width'] = max_header_width
            new_title_template['title_height'] = title_height
            new_title_template['header_height'] = header_height
            all_title_templates[i] = new_title_template

            if i > 2 and len(title_lines) <= last_lines_count:
                break

            last_lines_count = len(title_lines)

        lines_choice_max = i
        if box.has_top_icon:
            icon_size = int(box_theme.icon_size)
        else:
            icon_size = 0

        box_theme.save_title_templates(
            box._group_name, icon_size,
            all_title_templates[:lines_choice_max])

    # Now compare multiple possible areas for the box,
    # depending on BoxLayoutMode and number of lines for the box title.

    layout_mode = box._layout_mode
    BoxLayout.init_from_box(box, ports_min_sizes)
    box_layouts = list[BoxLayout]()

    if box.current_port_mode in (PortMode.INPUT, PortMode.OUTPUT):
        for i in range(1, lines_choice_max + 1):
            box_layouts.append(
                BoxLayout(i, BoxLayoutMode.LARGE,
                            TitleOn.SIDE, all_title_templates[i]))

        if box.has_top_icon:
            for i in range(1, lines_choice_max + 1):
                box_layouts.append(
                    BoxLayout(i, BoxLayoutMode.LARGE,
                                TitleOn.SIDE_UNDER_ICON,
                                all_title_templates[i]))

        for i in range(1, lines_choice_max + 1):
            box_layouts.append(
                BoxLayout(i, BoxLayoutMode.HIGH,
                            TitleOn.TOP, all_title_templates[i]))
    else:
        for i in range(1, lines_choice_max + 1):
            box_layouts.append(
                BoxLayout(i, BoxLayoutMode.LARGE,
                            TitleOn.TOP, all_title_templates[i]))

        for i in range(1, lines_choice_max + 1):
            box_layouts.append(
                BoxLayout(i, BoxLayoutMode.HIGH,
                            TitleOn.TOP, all_title_templates[i]))

    # sort areas and choose the first one (the littlest area)
    box_layouts.sort()

    high_layout, large_layout = None, None

    for layout in box_layouts:
        if (high_layout is None
                and layout.layout_mode is BoxLayoutMode.HIGH):
            high_layout = layout
        elif (large_layout is None
                and layout.layout_mode is BoxLayoutMode.LARGE):
            large_layout = layout

    if high_layout is None or large_layout is None:
        # can not happen, just for typing
        raise Exception

    high_layout.set_choosed()
    large_layout.set_choosed()

    if layout_mode is BoxLayoutMode.AUTO:
        if high_layout is box_layouts[0]:
            return high_layout, large_layout
        return large_layout, high_layout

    if layout_mode is BoxLayoutMode.HIGH:
        return high_layout, large_layout

    return large_layout, high_layout

def _set_ports_y_positions(
        box: 'BoxWidget',
        align_port_types: bool) -> dict[str, list[list[float]]]:
    '''ports Y positioning, return height segments info
    used if port-in-offset or port-out-offset are not zero in box theme'''

    def set_widget_pos(widget: QGraphicsItem, pos: float):
        match box._wrapping_state:          
            case WrappingState.WRAPPING:
                widget.setY(pos - ((pos - wrapped_port_pos)
                                    * box._wrapping_ratio))
            case WrappingState.UNWRAPPING:
                widget.setY(wrapped_port_pos + ((pos - wrapped_port_pos)
                                                * box._wrapping_ratio))
            case WrappingState.WRAPPED:
                widget.setY(wrapped_port_pos)
            case _:
                widget.setY(pos)

    if box._layout is None:
        raise Exception('._layout is needed')

    box_theme = box.get_theme()
    port_spacing = box_theme.port_spacing
    port_type_spacing = box_theme.port_type_spacing
    pen_width = box_theme.border_width
    last_in_type_and_sub = (PortType.NULL, PortSubType.REGULAR)
    last_out_type_and_sub = (PortType.NULL, PortSubType.REGULAR)
    last_type_and_sub = (PortType.NULL, PortSubType.REGULAR)

    port_type_spacing_in = port_type_spacing_out = port_type_spacing
    one_column = box._layout.one_column

    if box.has_side_title:
        start_pos = pen_width + port_spacing
    else:
        start_pos = pen_width + box._header_height

    if box._layout.title_on is TitleOn.TOP:
        wrapped_port_pos = (box._layout.wrapped_height
                            - canvas.theme.port_height - pen_width)
    else:
        wrapped_port_pos = pen_width + port_spacing

    last_in_pos = last_out_pos = start_pos

    # manage exceedent height in the box
    # due to the grid height (or to the others side ports
    # in a grouped box).
    exc_in = box._layout.exceeding_y_ins
    exc_out = box._layout.exceeding_y_outs
    exc_inout = box._layout.exceeding_y_inouts
    n_types_in = box._layout._pms.n_in_type_and_subs
    n_types_out = box._layout._pms.n_out_type_and_subs
    n_types_inout = box._layout._pms.n_inout_type_and_subs

    if one_column:
        new_pts = exc_inout / (1 + n_types_inout)
        if new_pts > port_type_spacing:
            port_type_spacing = port_type_spacing_in = \
                port_type_spacing_out = new_pts
            last_in_pos += new_pts
            last_out_pos += new_pts
        else:
            last_in_pos += exc_inout / 2
            last_out_pos += exc_inout / 2

    elif align_port_types:
        exceeding = min(exc_in, exc_out)
        n_types = max(n_types_in, n_types_out)
        new_pts = exceeding / (1 + n_types)
        if new_pts > port_type_spacing:
            port_type_spacing = new_pts
            port_type_spacing_in = port_type_spacing_out = new_pts
            last_in_pos += port_type_spacing
            last_out_pos += port_type_spacing
        else:
            last_in_pos += exceeding / 2
            last_out_pos += exceeding / 2

    elif box.current_port_mode is PortMode.BOTH:
        if exc_out < exc_in:
            new_pts_out = exc_out / (1 + n_types_out)
            more_out = 0.0
            if new_pts_out > port_type_spacing:
                port_type_spacing_out = new_pts_out
                last_in_pos += new_pts_out
                last_out_pos += new_pts_out
                more_out = (new_pts_out - port_type_spacing) * (n_types_out -1)

            if n_types_in > 1:
                port_type_spacing_in += (
                    (exc_in + more_out - exc_out) / (n_types_in - 1))
        else:
            new_pts_in = exc_in / (1 + n_types_in)
            more_in = 0.0
            if new_pts_in > port_type_spacing:
                port_type_spacing_in = new_pts_in
                last_in_pos += new_pts_in
                last_out_pos += new_pts_in
                more_in = (new_pts_in - port_type_spacing) * (n_types_in - 1)

            if n_types_out > 1:
                port_type_spacing_out += (
                    (exc_out + more_in - exc_in) / (n_types_out - 1))
    else:
        new_pts_in = exc_in / (n_types_in + 1)
        if new_pts_in > port_type_spacing_in:
            port_type_spacing_in = new_pts_in
            last_in_pos += new_pts_in
        else:
            last_in_pos += exc_in / 2

        new_pts_out = exc_out / (n_types_out + 1)
        if new_pts_out > port_type_spacing_out:
            port_type_spacing_out = new_pts_out
            last_out_pos += new_pts_out
        else:
            last_out_pos += exc_out / 2

    input_segments = list[list[float]]()
    output_segments = list[list[float]]()
    in_segment = [last_in_pos, last_in_pos]
    out_segment = [last_out_pos, last_out_pos]

    for port_type, port_subtype in list_port_types_and_subs():
        for port in box._port_list:
            if (port.port_type is not port_type
                    or port.port_subtype is not port_subtype):
                continue

            if one_column:
                last_in_pos = last_out_pos = max(last_in_pos, last_out_pos)

            if port.portgrp_id and port.pg_pos > 0:
                continue

            type_and_sub = (port.port_type, port.port_subtype)
            if one_column:
                if type_and_sub != last_type_and_sub:
                    if last_type_and_sub != (PortType.NULL, PortSubType.REGULAR):
                        last_in_pos += port_type_spacing
                        last_out_pos += port_type_spacing
                    last_type_and_sub = type_and_sub

            if port.port_mode is PortMode.INPUT:
                if not one_column and type_and_sub != last_in_type_and_sub:
                    if last_in_type_and_sub != (PortType.NULL, PortSubType.REGULAR):
                        last_in_pos += port_type_spacing_in
                    last_in_type_and_sub = type_and_sub

                if last_in_pos >= in_segment[1] + port_spacing + port_type_spacing_in:
                    if in_segment[0] != in_segment[1]:
                        input_segments.append(in_segment)
                    in_segment = [last_in_pos, last_in_pos]

                if port.portgrp_id:
                    # we place the portgroup widget and all its ports now
                    # because in one column mode, we can't be sure
                    # that port consecutivity isn't break by a port with
                    # another mode:
                    #
                    # input L
                    #     output L
                    # input R
                    #     output R
                    portgrp = port.portgrp
                    if portgrp is not None:
                        if portgrp.widget is not None:
                            set_widget_pos(portgrp.widget, last_in_pos)

                        for gp_port in portgrp.ports:
                            set_widget_pos(gp_port.widget, last_in_pos)
                            last_in_pos += canvas.theme.port_height
                else:
                    set_widget_pos(port.widget, last_in_pos)
                    last_in_pos += canvas.theme.port_height
                in_segment[1] = last_in_pos
                last_in_pos += port_spacing

            elif port.port_mode is PortMode.OUTPUT:
                if not one_column and type_and_sub != last_out_type_and_sub:
                    if last_out_type_and_sub != (PortType.NULL, PortSubType.REGULAR):
                        last_out_pos += port_type_spacing_out
                    last_out_type_and_sub = type_and_sub

                if last_out_pos >= out_segment[1] + port_spacing + port_type_spacing_out:
                    if out_segment[0] != out_segment[1]:
                        output_segments.append(out_segment)
                    out_segment = [last_out_pos, last_out_pos]

                if port.portgrp_id:
                    portgrp = port.portgrp
                    if portgrp is not None:
                        if portgrp.widget is not None:
                            set_widget_pos(portgrp.widget, last_out_pos)

                        for gp_port in portgrp.ports:
                            set_widget_pos(gp_port.widget, last_out_pos)
                            last_out_pos += canvas.theme.port_height
                else:
                    set_widget_pos(port.widget, last_out_pos)
                    last_out_pos += canvas.theme.port_height

                out_segment[1] = last_out_pos
                last_out_pos += port_spacing

        if align_port_types:
            # align port types horizontally
            if last_in_pos > last_out_pos:
                last_out_type_and_sub = last_in_type_and_sub
            else:
                last_in_type_and_sub = last_out_type_and_sub
            last_in_pos = last_out_pos = max(last_in_pos, last_out_pos)

    if in_segment[0] != in_segment[1]:
        input_segments.append(in_segment)
    if out_segment[0] != out_segment[1]:
        output_segments.append(out_segment)

    ports_top_in = 0.0
    ports_bottom_in = 0.0
    ports_top_out = 0.0
    ports_bottom_out = 0.0

    if input_segments:
        ports_top_in = input_segments[0][0]
        ports_bottom_in = input_segments[-1][1]

    if output_segments:
        ports_top_out = output_segments[0][0]
        ports_bottom_out = output_segments[-1][1]

    wp_ports_bottom = wrapped_port_pos + canvas.theme.port_height

    match box._wrapping_state:
        case WrappingState.WRAPPED:
            ports_top_in = ports_top_out = wrapped_port_pos
            ports_bottom_in = ports_bottom_out = wp_ports_bottom

        case WrappingState.WRAPPING:
            ports_top_in = from_float_to(
                ports_top_in, wrapped_port_pos, box._wrapping_ratio)
            ports_top_out = from_float_to(
                ports_top_out, wrapped_port_pos, box._wrapping_ratio)
            ports_bottom_in = from_float_to(
                ports_bottom_in, wp_ports_bottom, box._wrapping_ratio)
            ports_bottom_out = from_float_to(
                ports_bottom_out, wp_ports_bottom, box._wrapping_ratio)

        case WrappingState.UNWRAPPING:
            ports_top_in = from_float_to(
                wrapped_port_pos, ports_top_in, box._wrapping_ratio)
            ports_top_out = from_float_to(
                wrapped_port_pos, ports_top_out, box._wrapping_ratio)
            ports_bottom_in = from_float_to(
                wp_ports_bottom, ports_bottom_in, box._wrapping_ratio)
            ports_bottom_out = from_float_to(
                wp_ports_bottom, ports_bottom_out, box._wrapping_ratio)

    box._layout.set_ports_top_bottom(
        ports_top_in, ports_bottom_in,
        ports_top_out, ports_bottom_out)

    return {'input_segments': input_segments,
            'output_segments': output_segments}

def _set_ports_x_positions(box: 'BoxWidget', ports_min_sizes: PortsMinSizes):
    box_theme = box.get_theme()
    port_in_offset = box_theme.port_in_offset
    port_out_offset = box_theme.port_out_offset

    max_in_width = ports_min_sizes.ins_width
    max_out_width = ports_min_sizes.outs_width

    # Horizontal ports re-positioning
    in_x = port_in_offset
    out_x = box._width - max_out_width

    # Horizontal ports not in portgroup re-positioning
    for port in box._port_list:
        if port.portgrp_id:
            continue

        if port.port_mode is PortMode.INPUT:
            port.widget.setX(in_x)
            port.widget.set_port_width(max_in_width - port_in_offset)
        elif port.port_mode is PortMode.OUTPUT:
            port.widget.setX(out_x)
            port.widget.set_port_width(max_out_width - port_out_offset)

    # Horizontal portgroups and ports in portgroup re-positioning
    for portgrp in box._portgrp_list:
        if portgrp.widget is not None:
            if portgrp.port_mode is PortMode.INPUT:
                portgrp.widget.set_portgrp_width(max_in_width - port_in_offset)
                portgrp.widget.setX(in_x)
            elif portgrp.port_mode is PortMode.OUTPUT:
                portgrp.widget.set_portgrp_width(max_out_width - port_out_offset)
                portgrp.widget.setX(out_x)

        max_port_in_pg_width = canvas.theme.port_grouped_width

        for port in portgrp.ports:
            if port.widget is not None:
                port_print_width = int(port.widget.get_text_width())

                # change port in portgroup width only if
                # portgrp will have a name
                # to ensure that portgroup widget is large enough
                max_port_in_pg_width = max(max_port_in_pg_width,
                                            port_print_width + 6)

        out_in_portgrp_x = (box._width - port_out_offset
                            - max_port_in_pg_width)

        if portgrp.widget is not None:
            portgrp.widget.set_ports_width(max_port_in_pg_width)

        for port in portgrp.ports:
            if port.widget is not None:
                port.widget.set_port_width(max_port_in_pg_width)
                if port.port_mode is PortMode.INPUT:
                    port.widget.setX(in_x)
                elif port.port_mode is PortMode.OUTPUT:
                    port.widget.setX(out_in_portgrp_x)

def _set_title_positions(box: 'BoxWidget'):
    '''set title lines, header lines and icon positions'''
    box._header_line_left = None
    box._header_line_right = None

    box_theme = box.get_theme()
    font_size = box_theme.font.pixelSize()
    font_spacing = int(font_size * 1.4)
    icon_size = box_theme.icon_size

    pen_width = box_theme.border_width

    header_theme = box.get_theme(BoxStyler.HEADER)
    if box.isSelected():
        header_theme = header_theme.selected
    mg = header_theme.margin
    
    # when client is client capable of gui state
    # gui button has margins
    if box._can_handle_gui:
        gui_button = canvas.theme.gui_button
        if box._gui_visible:
            gui_button = gui_button.gui_visible
        else:
            gui_button = gui_button.gui_hidden
        
        gui_margin = gui_button.margin
        mg += gui_margin
        mg += gui_button.border_width
    else:
        gui_margin = canvas.theme.margin_empty
    
    # get the rect where the title is placed (top, bottom, left, right)
    if box.has_side_title:
        if box.current_port_mode is PortMode.INPUT:
            left = box._width - pen_width - box._header_width + mg.ports_side
            right = box._width - pen_width - mg.free_side
        else:
            left = pen_width + mg.free_side
            right = pen_width + box._header_width - mg.ports_side

        top = pen_width + mg.top_side
        bottom = pen_width + box._header_height - mg.top_side
    else:
        left = pen_width + mg.sides
        right = box._width - pen_width - mg.sides
        top = pen_width + mg.top
        bottom = pen_width + box._header_height - mg.bottom

    # set title lines Y position
    if box._title_under_icon:
        title_y_start = top + icon_size + font_spacing + 3
    else:
        title_y_start = top + font_size + 1

    for i, title_line in enumerate(box._title_lines):
        title_line.y = title_y_start + i * font_spacing

    if not box._title_under_icon and len(box._title_lines) == 1:
        box._title_lines[0].y = int(
            top + (bottom - top - font_size * 0.6) * 0.5 + font_size * 0.6)

    if box.has_side_title:
        # In case the title is near to be vertically centered in the box
        # It's prettier to center it correctly
        if (box._title_lines
                and not box._title_under_icon
                and not box._can_handle_gui
                and box._title_lines[-1].y + int(font_size * 1.4) > box._height):
            y_correction = (box._height - box._title_lines[-1].y - 2) / 2 - 2
        else:
            y_correction = 0

        # set title lines and icon pos
        match box.current_port_mode:
            case PortMode.INPUT:
                for title_line in box._title_lines:
                    title_line.x = left + 4
                    title_line.y += y_correction

                box.set_top_icon_pos(
                    right - int(icon_size) - 3,
                    int(top + 3))

            case PortMode.OUTPUT:
                if box.has_top_icon and not box._title_under_icon:
                    for title_line in box._title_lines:
                        if title_line.y >= top + icon_size + 6:
                            title_line.x = left + 4
                        else:
                            title_line.x = left + 3 + icon_size + 3
                        title_line.y += y_correction
                else:
                    for title_line in box._title_lines:
                        title_line.x = left + 4
                        title_line.y += y_correction

                box.set_top_icon_pos(left + 3, top + 3)
        
        return

    # Now we are sure title is on top

    # get title global sizes
    max_title_size = 0
    max_title_icon_size = 0

    for title_line in box._title_lines:
        title_size = title_line.size
        if (box.has_top_icon
                and title_line.y <= top + icon_size + 6 + font_size):
            # title line is beside icon
            title_size += icon_size + 4
            max_title_icon_size = max(max_title_icon_size, title_size)
        max_title_size = max(max_title_size, title_size)

    # set title lines X position
    for title_line in box._title_lines:
        if (box.has_top_icon
                and title_line.y <= top + icon_size + 6 + font_size):
            # title line is beside the icon
            title_line.x = (left
                            + (right - left - max_title_icon_size) * 0.5
                            + icon_size + 4)
        else:
            title_line.x = left + (right - left - title_line.size) * 0.5

    # set icon position
    box.set_top_icon_pos(
        left + (right - left - max_title_icon_size) * 0.5, top + 3)

    # calculate header lines positions
    side_size = (box._width - max(max_title_icon_size, max_title_size)) * 0.5

    if side_size > 10:
        y = top + (bottom - top) * 0.5
        box._header_line_left = (5.0, y, side_size - 5.0, y)
        box._header_line_right = (box._width - side_size + 5.0, y,
                                    box._width - 5.0, y)

def _get_wrap_triangle_pos(box: 'BoxWidget') -> UnwrapButton:
    if box.has_side_title:
        if box._height - box._header_height >= 15.0:
            if box.current_port_mode is PortMode.OUTPUT:
                return UnwrapButton.LEFT
            else:
                return UnwrapButton.RIGHT

    if box._layout is None:
        raise Exception('_get_wrap_triangle_pos, _layout is needed')

    last_in_pos = box._layout.ports_bottom_in
    last_out_pos = box._layout.ports_bottom_out

    if box._height - box._header_height >= 64.0:
        if (box.current_port_mode is PortMode.BOTH
                and box._current_layout_mode is BoxLayoutMode.HIGH):
            if last_in_pos > last_out_pos:
                return UnwrapButton.RIGHT
            else:
                return UnwrapButton.LEFT

        elif box.current_port_mode is PortMode.INPUT:
            return UnwrapButton.RIGHT

        elif box.current_port_mode is PortMode.OUTPUT:
            return UnwrapButton.LEFT

        y_side_space = last_in_pos - last_out_pos

        if y_side_space < -10.0:
            return UnwrapButton.LEFT
        if y_side_space > 10.0:
            return UnwrapButton.RIGHT
        return UnwrapButton.CENTER

    return UnwrapButton.NONE

def update_positions(
        box: 'BoxWidget', even_animated=False, without_connections=False,
        scene_checks=True, theme_change=False, wrap_anim=False):
    '''Redraw the box, may take some time (~ 10ms for a 30 ports box).
    It checks the present ports and portgroups, choose the box size and
    set title and ports positions.

    even_animated : if we need to update the box even
    if the box is in animation.

    without_connections : optimization, if we redraw all groups, we can
    redraw connections after having redrawn all boxes (2 times faster).

    scene_checks : if we redraw multiple boxes, we can resize the scene
    and check box overlapping after having redrawn all boxes.

    theme_change : only when we change theme.

    wrap_anim : only while wrapping/unwrapping, does not check the
    present ports, size is based on saved sizes.'''

    if canvas.loading_items:
        return

    if (not (even_animated or wrap_anim)
            and box in canvas.scene.move_boxes):
        box.update_positions_pending = True
        # do not change box layout while box is moved by animation
        # update_positions will be called when animation is finished
        return

    box.prepareGeometryChange()

    if (box._wrapping_state
            in (WrappingState.NORMAL, WrappingState.WRAPPED)
            or even_animated):
        # update port/portgrp list if box is not in wrapping animation
        # or forced with even_animated
        box._current_port_mode = PortMode.NULL
        box._port_list.clear()
        box._portgrp_list.clear()

        for port in canvas.list_ports(group_id=box._group_id):
            if port.port_mode & box._port_mode:
                box._port_list.append(port)
                box._current_port_mode |= port.port_mode

        for portgrp in canvas.list_portgroups(group_id=box._group_id):
            if box._current_port_mode & portgrp.port_mode:
                box._portgrp_list.append(portgrp)

    if theme_change:
        if box.shadow is not None:
            box.shadow.set_theme(box.get_theme(BoxStyler.SHADOW))
        
        box.set_icon()

        for port in box._port_list:
            if port.hidden_conn_widget is not None:
                port.hidden_conn_widget.update_theme()
                port.hidden_conn_widget.update_line_gradient()

    if options.auto_hide_groups and not box._port_list:
        box.setVisible(False)
        return

    box.setVisible(True)

    align_port_types = _should_align_port_types(box)
    ports_min_sizes = _get_ports_min_sizes(box, align_port_types)

    if (box._wrapping_state in (WrappingState.NORMAL, WrappingState.WRAPPED)
            or even_animated):
        box_layout, alter_layout = _choose_box_layout(box, ports_min_sizes)
        box._layout = box_layout
        box._alter_layout = alter_layout
        box._current_layout_mode = box_layout.layout_mode
        box._title_under_icon = bool(
            box_layout.title_on is TitleOn.SIDE_UNDER_ICON)
        box._title_lines = _split_title(box, box_layout.n_lines)
        box._header_width = box_layout.header_width
        box._header_height = box_layout.header_height

        box._unwrapped_width = box_layout.width
        box._unwrapped_height = box_layout.height
        box._wrapped_width = box_layout.wrapped_width
        box._wrapped_height = box_layout.wrapped_height

    match box._wrapping_state:
        case WrappingState.NORMAL:
            box._width = box._unwrapped_width
            box._height = box._unwrapped_height

        case WrappingState.WRAPPED:
            box._width = box._wrapped_width
            box._height = box._wrapped_height

        case WrappingState.WRAPPING:
            box._width = (
                box._unwrapped_width
                - (box._unwrapped_width - box._wrapped_width)
                    * box._wrapping_ratio)
            box._height = (
                box._unwrapped_height
                - (box._unwrapped_height - box._wrapped_height)
                    * box._wrapping_ratio)

        case WrappingState.UNWRAPPING:
            box._width = (
                box._wrapped_width
                + (box._unwrapped_width - box._wrapped_width)
                    * box._wrapping_ratio)
            box._height = (
                box._wrapped_height
                + (box._unwrapped_height - box._wrapped_height)
                    * box._wrapping_ratio)

    ports_y_segments_dict = _set_ports_y_positions(box, align_port_types)

    _set_ports_x_positions(box, ports_min_sizes)
    _set_title_positions(box)

    if box._wrapping_state is WrappingState.NORMAL:
        box._wrap_triangle_pos = _get_wrap_triangle_pos(box)

    if (theme_change
            or box._width != box._ex_width
            or box._height != box._ex_height
            or ports_y_segments_dict != box._ex_ports_y_segments_dict):
        build_painter_path(box, ports_y_segments_dict)
        build_painter_path(box, ports_y_segments_dict, selected=True)

    if (box._width != box._ex_width
            or box._height != box._ex_height
            or box.scenePos() != box._ex_scene_pos):
        if scene_checks:
            canvas.scene.resize_the_scene()

    box._ex_width = box._width
    box._ex_height = box._height
    box._ex_ports_y_segments_dict = ports_y_segments_dict
    box._ex_scene_pos = box.scenePos()

    if not without_connections:
        box.repaint_lines(forced=True)

    if scene_checks:
        if (box._wrapping_state in (WrappingState.NORMAL,
                                        WrappingState.WRAPPED)
                and box.isVisible()):
            canvas.scene.deplace_boxes_from_repulsers([box])

    box.update_positions_pending = False

    box.update()

    # I do not understand why without this,
    # when a renamed port does not change the box geometry,
    # port and portgroups widgets aren't updated,
    # they will be if we click on it for exemple.
    box.update_ports()

def get_dummy_rect(box: 'BoxWidget') -> QRectF:
    '''Used only for dummy box, to know its size
    before joining or arranging.'''
    box._current_port_mode = PortMode.NULL
    box._port_list.clear()
    box._portgrp_list.clear()

    for port in canvas.list_ports(group_id=box._group_id):
        if port.port_mode & box._port_mode:
            box._port_list.append(port)
            box._current_port_mode |= port.port_mode

    for portgrp in canvas.list_portgroups(group_id=box._group_id):
        if box._current_port_mode & portgrp.port_mode:
            box._portgrp_list.append(portgrp)

    align_port_types = _should_align_port_types(box)
    ports_min_sizes = _get_ports_min_sizes(box, align_port_types)
    box_layout, alter_layout = _choose_box_layout(box, ports_min_sizes)

    if box.is_hardware:
        hwr = float(canvas.theme.hardware_rack_width)
    else:
        hwr = 0.0

    if box.is_wrapped:
        return QRectF(-hwr, -hwr, box_layout.full_wrapped_width,
                        box_layout.full_wrapped_height)

    return QRectF(-hwr, -hwr,
                    box_layout.full_width, box_layout.full_height)

def get_layout(box: 'BoxWidget',
               layout_mode: BoxLayoutMode | None =None) -> BoxLayout:
    if box._layout is None:
        raise Exception('get_layout, ._layout is required !')

    if layout_mode is None:
        return box._layout

    if layout_mode is BoxLayoutMode.LARGE:
        if box._current_layout_mode is BoxLayoutMode.LARGE:
            return box._layout
        if box._alter_layout is None:
            raise Exception('get_layout, ._alter_layout is required !')
        return box._alter_layout

    if layout_mode is BoxLayoutMode.HIGH:
        if box._current_layout_mode is BoxLayoutMode.HIGH:
            return box._layout
        if box._alter_layout is None:
            raise Exception('get_layout, ._alter_layout is required !')
        return box._alter_layout

    return box._layout


from pathlib import Path

from .style_attributer import StyleAttributer


class UslStyleAttributer(StyleAttributer):
    def __init__(self, path: str, parent=None):
        super().__init__(path, parent=parent)
        self.selected = StyleAttributer(path + '.selected', self)
        self.subs.append('selected')


class ClientBoxStyleAttributer(UslStyleAttributer):
    def __init__(self, path: str, parent=None):
        super().__init__(path, parent)
        self.with_gui = UslStyleAttributer(path + '.with_gui', self)
        self.subs.append('with_gui')


class BoxStyleAttributer(UslStyleAttributer):
    def __init__(self, path: str, parent):
        super().__init__(path, parent)
        self.hardware = UslStyleAttributer(path + '.hardware', self)
        self.client = ClientBoxStyleAttributer(path + '.client', self)
        self.monitor = UslStyleAttributer(path + '.monitor', self)
        self.subs += ['hardware', 'client', 'monitor']


class PortStyleAttributer(UslStyleAttributer):
    def __init__(self, path: str, parent):
        super().__init__(path, parent)
        self.audio = UslStyleAttributer(path + '.audio', self)
        self.midi = UslStyleAttributer(path + '.midi', self)
        self.cv = UslStyleAttributer(path + '.cv', self)
        self.alsa = UslStyleAttributer(path + '.alsa', self)
        self.video = UslStyleAttributer(path + '.video', self)
        self.subs += ['audio', 'midi', 'cv', 'video', 'alsa']


class LineStyleAttributer(UslStyleAttributer):
    def __init__(self, path: str, parent):
        super().__init__(path, parent)
        self.audio = UslStyleAttributer(path + '.audio', self)
        self.midi = UslStyleAttributer(path + '.midi', self)
        self.alsa = UslStyleAttributer(path + '.alsa', self)
        self.video = UslStyleAttributer(path + '.video', self)
        self.disconnecting = StyleAttributer(path + '.disconnecting', self)
        self.subs += ['audio', 'midi', 'alsa', 'video', 'disconnecting']


class GuiButtonStyleAttributer(StyleAttributer):
    def __init__(self, path: str, parent):
        super().__init__(path, parent)
        self.gui_visible = StyleAttributer('.gui_visible', self)
        self.gui_hidden = StyleAttributer('.gui_hidden', self)
        self.subs += ['gui_visible', 'gui_hidden']


class GridStyleAttributer(StyleAttributer):
    def __init__(self, path: str, parent=None):
        super().__init__(path, parent)
        self._grid_min_width = 100.0
        self._grid_min_height = 100.0

        self.technical_grid = StyleAttributer('.technical_grid', self)
        self.grid = StyleAttributer('.grid', self)
        self.chessboard = StyleAttributer('.chessboard', self)
        self.subs += ['technical_grid', 'grid', 'chessboard']


class IconTheme:
    def __init__(self):
        src = ':/canvas/dark/'
        self.hardware_capture = src + 'microphone.svg'
        self.hardware_playback = src + 'audio-headphones.svg'
        self.hardware_grouped = src + 'pb_hardware.svg'
        self.hardware_midi = src + 'DIN-5.svg'
        self.monitor_capture = src + 'monitor_capture.svg'
        self.monitor_playback = src + 'monitor_playback.svg'

    def read_theme(self, theme_file: Path):
        icons_dir = theme_file.parent / 'icons'
        if not icons_dir.is_dir():
            return

        for key in ('hardware_capture', 'hardware_playback', 'hardware_grouped',
                    'hardware_midi', 'monitor_capture', 'monitor_playback'):
            icon_path = icons_dir / f'{key}.svg'
            if icon_path.is_file():
                self.__setattr__(key, str(icon_path))




from enum import Enum, auto

from qtpy.QtGui import QFont

from ..theme import StyleAttributer


class UnwrapButton(Enum):
    NONE = 0
    LEFT = 1
    CENTER = 2
    RIGHT = 3


class WrappingState(Enum):
    NORMAL = 0
    WRAPPING = 1
    WRAPPED = 2
    UNWRAPPING = 3


class TitleLine:
    text = ''
    size = 0.0
    x = 0.0
    y = 0.0
    is_little = False

    def __init__(self, text: str, theme: StyleAttributer, little=False):
        self.theme = theme
        self.text = text
        self.is_little = little
        self.x = 0.0
        self.y = 0.0

        self.font = None
        self.size = theme.get_text_width(text)

    def get_font(self) -> QFont:
        return self.theme.font


class PaintElement(Enum):
    MAIN = auto()
    HEADER = auto()
    ANTI_HEADER = auto()
    GUI_BUTTON = auto()


class BoxStyler(Enum):
    BOX = auto()
    SHADOW = auto()
    HEADER = auto()
    HEADER_LINE = auto()
    WRAPPER = auto()
    PORTS_BORDER = auto()


import logging
import time

from qtpy.QtCore import Signal, Slot, QTimer, QObject # type:ignore

from patshared import PortMode

from .init_values import AliasingReason, canvas
from .grouped_lines_widget import GroupedLinesWidget


_logger = logging.getLogger(__name__)


def _join_group(group_id: int):
    '''join group boxes once animation is finished'''
    group = canvas.get_group(group_id)
    if group is None:
        _logger.error(
            f"join_group({group_id}) - unable to find groups to join")
        return

    if not group.splitted:
        _logger.error(f"join_group({group_id}) - group is not splitted")
        return

    wrap = True
    for box in group.widgets:
        wrap = wrap and box.is_wrapped

    eater, eaten = group.widgets

    for portgroup in canvas.list_portgroups(group_id=group_id):
        if (portgroup.port_mode is eaten.port_mode
                and portgroup.widget is not None):
            portgroup.widget.setParentItem(eater)

    for port in canvas.list_ports(group_id=group_id):
        if (port.port_mode is eaten.port_mode
                and port.widget is not None):
            port.widget.setParentItem(eater)

    eater.set_port_mode(PortMode.BOTH)
    eaten.remove_icon_from_scene()
    canvas.scene.remove_box(eaten)
    group.widgets.remove(eaten)
    canvas.remove_box(eaten)
    group.splitted = False
    del eaten

    eater.send_move_callback()
    eater.set_wrapped(wrap, animate=False)
    eater.update_positions(scene_checks=False)

    canvas.cb.group_joined(group_id)
    QTimer.singleShot(0, canvas.scene.update)


class CanvasObject(QObject):
    move_boxes_finished = Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self._gps_to_join = set[int]()
        self.move_boxes_finished.connect(self._join_after_move)

        self.connect_update_timer = QTimer()
        self.connect_update_timer.setInterval(0)
        self.connect_update_timer.setSingleShot(True)
        self.connect_update_timer.timeout.connect(
            self._connect_update_timer_finished)

        self._aliasing_reason = AliasingReason.NONE
        self._aliasing_timer_started_at = 0.0
        self._aliasing_move_timer = QTimer()
        self._aliasing_move_timer.setInterval(0)
        self._aliasing_move_timer.setSingleShot(True)
        self._aliasing_move_timer.timeout.connect(
            self._aliasing_move_timer_finished)

        self._aliasing_view_timer = QTimer()
        self._aliasing_view_timer.setInterval(500)
        self._aliasing_view_timer.setSingleShot(True)
        self._aliasing_view_timer.timeout.connect(
            self._aliasing_view_timer_finished)

    @Slot()
    def _connect_update_timer_finished(self):
        GroupedLinesWidget.change_all_prepared_conns()

    @Slot()
    def _aliasing_move_timer_finished(self):
        if time.time() - self._aliasing_timer_started_at > 0.060:
            canvas.set_aliasing_reason(self._aliasing_reason, True)

        if self._aliasing_reason is AliasingReason.VIEW_MOVE:
            self._aliasing_view_timer.start()

    @Slot()
    def _aliasing_view_timer_finished(self):
        canvas.set_aliasing_reason(AliasingReason.VIEW_MOVE, False)

    def start_aliasing_check(self, aliasing_reason: AliasingReason):
        self._aliasing_reason = aliasing_reason
        self._aliasing_timer_started_at = time.time()
        self._aliasing_move_timer.start()

    @Slot()
    def _join_after_move(self):
        for group_id in self._gps_to_join:
            _join_group(group_id)

        self._gps_to_join.clear()
        canvas.cb.animation_finished()

    def add_group_to_join(self, group_id: int):
        self._gps_to_join.add(group_id)

    def rm_group_to_join(self, group_id: int):
        self._gps_to_join.discard(group_id)

    def rm_all_groups_to_join(self):
        self._gps_to_join.clear()



#!/usr/bin/python3
import logging
from pathlib import Path
from typing import Any

from qtpy.QtCore import Qt
from qtpy.QtGui import QColor, QFont, QImage, QFontDatabase

from . import theme_cache
from .theme_utils import to_qcolor, ThemeFile
from .theme_structs import BorderMode, Align
from .style_attributer import StyleAttributer
from .style_attributers import (
    BoxStyleAttributer, PortStyleAttributer, LineStyleAttributer,
    UslStyleAttributer, GuiButtonStyleAttributer,
    GridStyleAttributer, IconTheme)


_logger = logging.getLogger(__name__)

    
_DEFAULT_STYLE_ATTRS = {
    'background': QColor('black'),
    'background2': QColor(),
    'background-image': QImage(),
    'border-color': QColor('white'),
    'border-mode': BorderMode.DEFAULT,
    'border-radius': 0,
    'border-style': Qt.PenStyle.SolidLine,
    'border-width': 1,
    'font-name': "Deja Vu Sans",
    'font-size': 11,
    'font-weight': QFont.Weight.Normal,
    'grid-min-width': 100,
    'grid-min-height': 100,
    'icon-size': 24,
    'margin-bottom': 3,
    'margin-free-side': 3,
    'margin-ports-side': 3,
    'margin-sides': 3,
    'margin-top': 3,
    'margin-top-side': 3,
    'padding-bottom': 3,
    'padding-free-side': 3,
    'padding-ports-side': 3,
    'padding-sides': 3,
    'padding-top': 3,
    'padding-top-side': 3,
    'output-align': Align.LEFT,
    'port-in-offset': 0,
    'port-in-offset-mode': 'bore',
    'port-out-offset': 0,
    'port-out-offset-mode': 'bore',
    'port-spacing': 2,
    'port-type-spacing': 2,
    'text-color': QColor('white'),
    'visible': False,
    'drilled': False
}


class Theme(StyleAttributer):
    theme_file_path = Path()

    def __init__(self):
        super().__init__('')

        # fallbacks values for all (ugly style, but better than nothing)
        self._attrs = _DEFAULT_STYLE_ATTRS

        self.scene_background_color = QColor('black')
        self.scene_background_image: QImage | None = None
        self.monitor_color = QColor(190, 158, 0)
        self.port_height = 16

        self.port_grouped_width = 19
        self.box_spacing = 4
        self.box_spacing_horizontal = 24
        self.hardware_rack_width = 5
        self.port_type_colors = dict[str, QColor]()

        self.icon = IconTheme()

        self.aliases = dict[str, Any]()

        self.box = BoxStyleAttributer('.box', self)
        self.box_wrapper = BoxStyleAttributer('.box_wrapper', self)
        self.box_header_line = BoxStyleAttributer('.box_header_line', self)
        self.box_shadow = BoxStyleAttributer('.box_shadow', self)
        self.box_header = BoxStyleAttributer('.box_header', self)
        self.box_ports_border = BoxStyleAttributer('.box_ports_border', self)
        self.portgroup = PortStyleAttributer('.portgroup', self)
        self.port = PortStyleAttributer('.port', self)
        self.line = LineStyleAttributer('.line', self)
        self.rubberband = StyleAttributer('.rubberband', self)
        self.hardware_rack = UslStyleAttributer('.hardware_rack', self)
        self.monitor_decoration = UslStyleAttributer(
            '.monitor_decoration', self)
        self.gui_button = GuiButtonStyleAttributer('.gui_button', self)
        self.grid = GridStyleAttributer('.grid', self)

        self.subs += [
            'box', 'box_wrapper', 'box_header_line', 'box_shadow',
            'box_header', 'box_ports_border', 'portgroup', 'port', 'line',
            'rubberband', 'hardware_rack', 'monitor_decoration',
            'gui_button', 'grid']

    @classmethod
    def set_file_path(cls, theme_file_path: Path):
        ThemeFile.path = theme_file_path

    @classmethod
    def load_cache(cls):
        theme_cache.load()

    @classmethod
    def save_cache(cls):
        theme_cache.save()

    def _read_body_attr(self, body_key: str, body_value):
        match body_key:
            case 'port-height'|'box-spacing-horizontal' \
                    |'hardware-rack-width' | 'port-grouped-width':
                if not isinstance(body_value, int):
                    return
                self.__setattr__(body_key.replace('-', '_'), body_value)

            case 'box-spacing':
                # box_spacing must be an even number
                if not isinstance(body_value, int):
                    return
                self.box_spacing = 2 * (body_value // 2)

            case 'background':
                scene_bg_color = to_qcolor(body_value)
                if scene_bg_color is None:
                    scene_bg_color = QColor('black')
                self.scene_background_color = scene_bg_color

            case 'background-image':
                if not isinstance(body_value, str):
                    return

                background_path = \
                    ThemeFile.path.parent / 'images' / body_value
                if background_path.is_file():
                    try:
                        self.scene_background_image = QImage(
                            str(background_path))
                        if self.scene_background_image.isNull():
                            _logger.error(
                                f"background {background_path} "
                                "is not a valid image")
                            self.scene_background_image = None
                    except:
                        _logger.error(
                            f"background {background_path} "
                            "is not a valid image")
                else:
                    _logger.error(
                        "Unable to find background-image "
                        f"\"{background_path}\"")

            case 'monitor-color':
                monitor_color = to_qcolor(body_value)
                if monitor_color is None:
                    monitor_color = QColor(190, 158, 0)
                self.monitor_color = monitor_color
                
            case _:
                _logger.warning(
                    f'Theme [body]{body_key} is unknown '
                    'and will have no effect')

    def read_theme(self, theme_dict: dict[str, dict], theme_file_path: Path,
                   for_linter=False):
        '''theme_file_path is only used here to find external resources'''
        _logger.info(f'start to read theme {theme_dict}, {theme_file_path}')
        if not isinstance(theme_dict, dict):
            _logger.error("invalid dict read error")
            return

        # set some specific default values
        self.box_header.set_attribute('margin', 0.0)
        self.gui_button._attrs['text-color'] = QColor()

        ThemeFile.path = theme_file_path
        self.icon.read_theme(theme_file_path)

        if not for_linter:
            # install all fonts from theme 'fonts' directory
            fonts_dir = Path(theme_file_path).parent / 'fonts'
            if fonts_dir.is_dir():
                for font_path in fonts_dir.iterdir():
                    if str(font_path).endswith(('.otf', '.ttf')):
                        try:
                            QFontDatabase.addApplicationFont(str(font_path))
                        except:
                            _logger.warning(
                                "failed to install font "
                                f"from file {str(font_path)}")

        self.aliases.clear()

        # first read if there are any aliases
        for key, value in theme_dict.items():
            if key != 'aliases':
                continue

            port_types = ('audio', 'midi', 'cv', 'alsa', 'video')

            if not isinstance(value, dict):
                _logger.error(f"'{key}' must contains a dictionnary, ignored")
                continue

            for alias_key, alias_value in value.items():
                if not isinstance(alias_key, str):
                    _logger.error(
                        "alias key must be a string. "
                        f"Ignore: {str(alias_key)}")
                    continue

                self.aliases[alias_key] = str(alias_value)
                if alias_key in port_types:
                    port_color = to_qcolor(alias_value)
                    if port_color is None:
                        port_color = QColor('black')
                    port_color = port_color.toRgb()
                    self.port_type_colors[alias_key] = port_color

            for port_type in port_types:
                if self.port_type_colors.get(port_type) is None:
                    _logger.error(f"An alias named '{port_type}' is needed")
                    self.port_type_colors[port_type] = QColor()

            break

        # read and parse the dict
        for key, value in theme_dict.items():
            if key in ('aliases', 'Theme'):
                continue

            begin, point, end = key.partition('.')

            if not isinstance(value, dict):
                _logger.error(f"'{key}' must contains a dictionnary, ignored")
                continue

            if begin not in ['body'] + self.subs:
                _logger.error(f"invalid ignored block key: [{key}]")
                continue

            # replace alias with alias value
            for sub_key, sub_value in value.items():
                if not isinstance(sub_value, str):
                    continue

                for alias_key, alias_value in self.aliases.items():
                    if alias_key not in sub_value:
                        continue

                    if sub_value == alias_key:
                        value[sub_key] = alias_value
                        break

                    new_words = list[str]()

                    for word in sub_value.split(' '):
                        if word == alias_key:
                            new_words.append(alias_value)
                        elif word == f'-{alias_key}':
                            new_words.append(f'-{alias_value}')
                        else:
                            new_words.append(word)

                    value[sub_key] = ' '.join(new_words)

            if key == 'body':
                for body_key, body_value in value.items():
                    self._read_body_attr(body_key, body_value)
                continue

            inherits_name = value.get('inherits')
            if isinstance(inherits_name, str):
                this = self.child(key)
                mother = self.child(inherits_name)

                if this is None:
                    # should not happen
                    _logger.error(f'[{key}] not found to inherit')                
                elif mother is None:
                    _logger.warning(
                        f'[{key}]{value}, {inherits_name} does not exists.')
                else:
                    _logger.info(f'{key} inherits {inherits_name}')
                    this.inherit(mother)

            sub_attributer = self.child(begin)
            if sub_attributer is not None:
                sub_attributer.set_style_dict(end, value)
        
        if self.scene_background_image is not None:
            if self.scene_background_color.alphaF() == 1.0:
                _logger.warning(
                    '[body]background-image will be ignored because '
                    '[body]background-color has no transparency')


import logging
from pathlib import Path
from typing import Callable, Iterator

from qtpy.QtGui import QCursor, QGuiApplication, QKeyEvent
from qtpy.QtWidgets import QMainWindow, QMessageBox, QWidget
from qtpy.QtCore import QTimer, QSettings, QThread, Qt

from patshared import (
    PortType, PortSubType, PortMode, JackMetadatas,
    PortTypesViewFlag, GroupPos, Naming, TransportPosition,
    ViewsDictEnsureOne, ViewData, PortgroupsDict, PortgroupMem, CustomNames)

from . import patchbay_batches, patchbay_hiddens, patchbay_views
from .bases.connection import Connection
from .bases.elements import ToolDisplayed, CanvasOptimizeIt, CanvasOptimize
from .bases.group import Group
from .bases.port import Port
from .bases.portgroup import Portgroup
from .calbacker import Callbacker
from .cancel_mng import CancelMng, CancelOp, CancellableAction
from .conns_clipboard import ConnClipboard
from .dialogs.options_dialog import CanvasOptionsDialog
from .menus.canvas_menu import CanvasMenu
from .patchbay_signals import SignalsObject
from .patchcanvas import patchcanvas
from .patchcanvas.utils import get_new_group_positions
from .patchcanvas.scene_view import PatchGraphicsView
from .patchcanvas.init_values import (
    AliasingReason, CanvasFeaturesObject, CanvasOptionsObject)
from .patchichi_export import export_to_patchichi_json
from .pretty_diff_checker import PrettyDiffChecker
from .tools_widgets import PatchbayToolsWidget
from .widgets.filter_frame import FilterFrame


_translate = QGuiApplication.translate
_logger = logging.getLogger(__name__)


def enum_to_flag(enum_int: int) -> int:
    if enum_int <= 0:
        return 0
    return 2 ** (enum_int - 1)

def in_main_thread():
    '''This decorator indicates that the decorated method will be executed
    directly if called in the main thread, otherwise as soon as possible
    in the main thread.'''
    def decorator(func: Callable):
        def wrapper(*args, **kwargs):
            mng: PatchbayManager = args[0]

            if QThread.currentThread() is QGuiApplication.instance().thread():
                return func(*args, **kwargs)

            mng.sg.to_main_thread.emit(func, args, kwargs)
        return wrapper
    return decorator


class PatchbayManager:
    naming = Naming.ALL
    port_types_view = PortTypesViewFlag.ALL

    canvas_optimize = CanvasOptimize.NORMAL

    groups = list[Group]()
    connections = list[Connection]()
    _groups_by_name = dict[str, Group]()
    _groups_by_id = dict[int, Group]()
    _ports_by_name = dict[str, Port]()
    _ports_by_uuid = dict[int, Port]()

    view_number = 1
    views = ViewsDictEnsureOne()

    portgroups_memory = PortgroupsDict()
    custom_names = CustomNames()

    jack_metadatas = JackMetadatas()

    def __init__(self, settings: QSettings):
        self.callbacker = Callbacker(self)
        self.cancel_mng = CancelMng(self)
        self._settings = settings

        self.main_win: QMainWindow | None = None
        self._tools_widget: PatchbayToolsWidget | None = None
        self.options_dialog: CanvasOptionsDialog | None = None
        self.filter_frame: FilterFrame | None = None

        self._manual_path: Path | None = None

        self.connections_clipboard = ConnClipboard(self)

        self.server_is_started = True

        self.sg = SignalsObject()

        self._next_group_id = 0
        self._next_port_id = 0
        self._next_connection_id = 0
        self._next_portgroup_id = 1

        self.client_uuids = dict[str, int]()
        '''Stock JACK client names and their uuid,
        uuid can be provided before the group creation.'''

        self.jack_export_naming = Naming.CUSTOM
        self.naming = Naming.from_config_str(settings.value(
            'Canvas/naming', 'ALL', type=str))
        self.pretty_diff_checker = PrettyDiffChecker(self)

        self.group_a2j_hw: bool = settings.value(
            'Canvas/group_a2j_ports', True, type=bool)
        self.alsa_midi_enabled: bool = settings.value(
            'Canvas/alsa_midi_enabled', False, type=bool)

        # all patchbay events are delayed
        # to reduce the patchbay comsumption.
        # Redraws in canvas are made once 50ms have passed without any event.
        # This prevent one group redraw per port added/removed
        # when a lot of ports are added/removed/renamed simultaneously.
        self._delayed_orders_timer = QTimer()
        self._delayed_orders_timer.setInterval(50)
        self._delayed_orders_timer.setSingleShot(True)
        self._delayed_orders_timer.timeout.connect(
            self._delayed_orders_timeout)

        self.sg.out_thread_order.connect(self._delayed_orders_timer.start)
        self.sg.to_main_thread.connect(self._execute_in_main_thread)

        self.sg.theme_changed.connect(self.change_theme)

        self.sg.a2j_grouped_changed.connect(self.set_a2j_grouped)
        self.sg.alsa_midi_enabled_changed.connect(self.set_alsa_midi_enabled)
        self.sg.group_shadows_changed.connect(self.set_group_shadows)
        self.sg.auto_select_items_changed.connect(self.set_auto_select_items)
        self.sg.elastic_changed.connect(self.set_elastic_canvas)
        self.sg.borders_nav_changed.connect(self.set_borders_navigation)
        self.sg.prevent_overlap_changed.connect(self.set_prevent_overlap)
        self.sg.max_port_width_changed.connect(patchcanvas.set_max_port_width)
        self.sg.default_zoom_changed.connect(patchcanvas.set_default_zoom)

        self._tools_displayed = ToolDisplayed.ALL

    def app_init(self,
                 view: PatchGraphicsView,
                 theme_paths: tuple[Path, ...],
                 manual_path: Path | None = None,
                 callbacker: Callbacker | None = None,
                 options: CanvasOptionsObject | None = None,
                 features: CanvasFeaturesObject | None = None,
                 default_theme_name='Black Gold'):
        if callbacker is None:
            self.callbacker = Callbacker(self)
        else:
            if not isinstance(callbacker, Callbacker):
                exception = TypeError(
                    "callbacker must be a Callbacker instance !")
                raise exception

            self.callbacker = callbacker

        self._manual_path = manual_path

        if options is None:
            options = patchcanvas.CanvasOptionsObject()
            if isinstance(self._settings, QSettings):
                options.set_from_settings(self._settings)

        if features is None:
            features = CanvasFeaturesObject()

        patchcanvas.set_options(options)
        patchcanvas.set_features(features)
        patchcanvas.init(
            view, self.callbacker, theme_paths, default_theme_name)
        patchcanvas.canvas.scene.scale_changed.connect(
            self._scene_scale_changed)

        # just to have the zoom slider updated with the default zoom
        patchcanvas.canvas.scene.zoom_reset()

    @property
    def very_fast_operation(self) -> bool:
        '''when True, items are not added/removed from patchcanvas.
        Useful to win time at startup or refresh'''
        return self.canvas_optimize is CanvasOptimize.VERY_FAST

    def _scene_scale_changed(self, value: float):
        self.sg.scene_scale_changed.emit(value)

    def _execute_in_main_thread(
            self, func: Callable, args: tuple, kwargs: dict):
        func(*args, **kwargs)

    # --- widgets related methods --- #

    def set_main_win(self, main_win: QWidget):
        self.main_win = main_win # type:ignore

    def set_tools_widget(self, tools_widget: PatchbayToolsWidget):
        self._tools_widget = tools_widget
        self._tools_widget.set_patchbay_manager(self)

    def set_canvas_menu(self, canvas_menu: CanvasMenu):
        self.canvas_menu = canvas_menu

    def set_filter_frame(self, filter_frame: FilterFrame):
        self.filter_frame = filter_frame
        self.filter_frame.set_settings(self._settings)

    def set_options_dialog(self, options_dialog: CanvasOptionsDialog):
        self.options_dialog = options_dialog

    def show_options_dialog(self):
        if self.options_dialog is None:
            return

        self.options_dialog.move(QCursor.pos())
        self.options_dialog.show()

    def view(self) -> ViewData:
        return self.views[self.view_number]

    # --- manager methods --- #

    @staticmethod
    def save_patchcanvas_cache():
        patchcanvas.save_cache()

    def save_settings(self):
        if self._settings is None:
            return

        self._settings.setValue('Canvas/naming', self.naming.name)
        self._settings.setValue('Canvas/group_a2j_ports',
                                self.group_a2j_hw)
        self._settings.setValue('Canvas/alsa_midi_enabled',
                                self.alsa_midi_enabled)
        patchcanvas.options.save_to_settings(self._settings)

    def optimize_operation(
            self, yesno: bool, auto_redraw=False, prevent_overlap=True):
        if patchcanvas.canvas is not None:
            patchcanvas.set_loading_items(
                yesno,
                auto_redraw=auto_redraw,
                prevent_overlap=prevent_overlap)

    def _add_group(self, group: Group):
        self.groups.append(group)
        self._groups_by_id[group.group_id] = group
        self._groups_by_name[group.name] = group

    def _remove_group(self, group: Group):
        if group in self.groups:
            self.groups.remove(group)
            self._groups_by_id.pop(group.group_id)
            self._groups_by_name.pop(group.name)

    def new_portgroup(self, group_id: int, port_mode: PortMode,
                      ports: tuple[Port, ...] | list[Port]) -> Portgroup:
        portgroup = Portgroup(self, group_id, self._next_portgroup_id,
                              port_mode, tuple(ports))
        self._next_portgroup_id += 1
        return portgroup

    def port_type_shown(
            self, full_port_type: tuple[PortType, PortSubType]) -> bool:
        port_type, sub_type = full_port_type

        match port_type:
            case PortType.MIDI_JACK:
                return PortTypesViewFlag.MIDI in self.port_types_view
            case PortType.AUDIO_JACK:
                match sub_type:
                    case PortSubType.REGULAR:
                        return PortTypesViewFlag.AUDIO in self.port_types_view
                    case PortSubType.CV:
                        return PortTypesViewFlag.CV in self.port_types_view
                    case (PortSubType.REGULAR | PortSubType.CV):
                        return (
                            (PortTypesViewFlag.AUDIO | PortTypesViewFlag.CV)
                            in self.port_types_view)
            case PortType.MIDI_ALSA:
                return (self.alsa_midi_enabled
                        and PortTypesViewFlag.ALSA in self.port_types_view)
            case PortType.VIDEO:
                return PortTypesViewFlag.VIDEO in self.port_types_view

        return False

    def animation_finished(self):
        '''Executed after any patchcanvas animation, it cleans
        in patchcanvas all boxes that should be hidden now.'''
        with CanvasOptimizeIt(self, auto_redraw=True, prevent_overlap=False):
            for group in self.groups:
                if group.current_position.hidden_port_modes() is PortMode.NULL:
                    continue

                hidden_port_mode = group.current_position.hidden_port_modes()

                if hidden_port_mode & PortMode.OUTPUT:
                    for conn in self.connections:
                        if conn.port_out.group_id is group.group_id:
                            conn.remove_from_canvas()

                if hidden_port_mode & PortMode.INPUT:
                    for conn in self.connections:
                        if conn.port_in.group_id is group.group_id:
                            conn.remove_from_canvas()

                for portgroup in group.portgroups:
                    if hidden_port_mode & portgroup.port_mode:
                        portgroup.remove_from_canvas()

                for port in group.ports:
                    if hidden_port_mode & port.mode:
                        port.remove_from_canvas()

                if group.outs_ptv is PortTypesViewFlag.NONE:
                    hidden_port_mode |= PortMode.OUTPUT
                if group.ins_ptv is PortTypesViewFlag.NONE:
                    hidden_port_mode |= PortMode.INPUT

                if hidden_port_mode is PortMode.BOTH:
                    group.remove_from_canvas()

            for conn in self.connections:
                conn.add_to_canvas()

        # patchcanvas.canvas.scene.resize_the_scene()
        self.sg.hidden_boxes_changed.emit()
        self.sg.animation_finished.emit()

    def set_group_hidden_sides(self, group_id: int, port_mode: PortMode):
        patchbay_hiddens.set_group_hidden_sides(self, group_id, port_mode)

    def restore_group_hidden_sides(
            self, group_id: int, scene_pos: tuple[int, int] | None =None):
        patchbay_hiddens.restore_group_hidden_sides(
            self, group_id, scene_pos=scene_pos)

    def restore_all_group_hidden_sides(self):
        patchbay_hiddens.restore_all_group_hidden_sides(self)

    def hide_all_groups(self):
        patchbay_hiddens.hide_all_groups(self)

    def list_hidden_groups(self) -> Iterator[Group]:
        for group in self.groups:
            if (group.current_position.hidden_port_modes()
                    is not PortMode.NULL):
                yield group

    def get_group_from_name(self, group_name: str) -> Group | None:
        return self._groups_by_name.get(group_name)

    def get_group_from_id(self, group_id: int) -> Group | None:
        return self._groups_by_id.get(group_id)

    def get_port_from_name(self, port_name: str) -> Port | None:
        return self._ports_by_name.get(port_name)

    def get_port_from_uuid(self, uuid:int) -> Port | None:
        return self._ports_by_uuid.get(uuid)

    def get_port_from_id(self, group_id: int, port_id: int) -> Port | None:
        group = self.get_group_from_id(group_id)
        if group is not None:
            for port in group.ports:
                if port.port_id == port_id:
                    return port

    def save_group_position(self, gpos: GroupPos):
        'reimplement this to save a group position elsewhere'
        pass

    def save_portgroup_memory(self, portgrp_mem: PortgroupMem):
        'reimplement this to save a portgroup memory elsewhere'
        pass

    def get_corrected_a2j_group_name(self, group_name: str) -> str:
        '''a2j replace points with spaces in the group name.
        we do nothing here, but in some conditions we can
        assume we know it.'''
        return group_name

    def set_group_as_nsm_client(self, group: Group):
        'do nothing, reimplement this'
        pass

    def get_group_position(self, group_name: str) -> GroupPos:
        ptv_view = self.view().ptvs.get(self.port_types_view)
        if ptv_view is None:
            ptv_view = dict[str, GroupPos]()
            self.view().ptvs[self.port_types_view] = ptv_view

        gpos = ptv_view.get(group_name)
        if gpos is not None:
            return gpos

        self.cancel_mng.new_pos_created = True

        is_white_list_view = self.view().is_white_list

        # prevent move to a new position in case of port_types_view change
        # if there is no remembered position for this group in new view
        if not is_white_list_view:
            group = self.get_group_from_name(group_name)
            if group is not None:
                # copy the group_position
                gpos = group.current_position.copy()
                gpos.port_types_view = self.port_types_view
                gpos.has_sure_existence = False
                gpos.set_hidden_port_mode(PortMode.NULL)
                ptv_view[group_name] = gpos
                self.save_group_position(gpos)
                return gpos

        # group position doesn't already exists, create one
        gpos = GroupPos()
        gpos.fully_set = False
        gpos.port_types_view = self.port_types_view
        gpos.group_name = group_name

        if is_white_list_view:
            gpos.set_hidden_port_mode(PortMode.BOTH)

        for port_mode, xy in get_new_group_positions().items():
            gpos.boxes[port_mode].pos = xy

        ptv_view[group_name] = gpos
        self.save_group_position(gpos)
        return gpos

    def add_portgroup_memory(self, portgroup_mem: PortgroupMem):
        self.portgroups_memory.save_portgroup(portgroup_mem)

        group = self.get_group_from_name(portgroup_mem.group_name)
        if group is not None:
            group.portgroup_memory_added(portgroup_mem)

    def remove_and_add_all(self):
        with CanvasOptimizeIt(self, auto_redraw=True):
            for connection in self.connections:
                connection.remove_from_canvas()

            for group in self.groups:
                for portgroup in group.portgroups:
                    portgroup.remove_from_canvas()

                for port in group.ports:
                    port.remove_from_canvas()

                group.remove_from_canvas()
                group.add_to_canvas()
                group.add_all_ports_to_canvas()

            for connection in self.connections:
                connection.add_to_canvas()

    def clear_all(self):
        patchcanvas.clear_all()
        self.connections.clear()
        self.jack_metadatas.clear()
        self.groups.clear()
        self._groups_by_id.clear()
        self._groups_by_name.clear()
        self._ports_by_name.clear()
        self._ports_by_uuid.clear()

        self._next_group_id = 0
        self._next_port_id = 0
        self._next_portgroup_id = 1
        self._next_connection_id = 0

        self.sg.all_groups_removed.emit()

    def save_view_and_port_types_view(self):
        pass

    def change_port_types_view(
            self, port_types_view: PortTypesViewFlag, force=False):
        patchbay_views.change_port_types_view(
            self, port_types_view, force=force)

    def set_views_changed(self):
        '''emit the `view_changed` signal. Can be Inherited to do other tasks'''
        self.sg.views_changed.emit()

    def new_view(self, view_number: int | None =None,
                 exclusive_with: dict[int, PortMode] | None =None):
        '''create a new view and switch directly to this view.

        If `view_number` is not set, it will choose the first available
        number.

        if `exclusive_with` is set, all non matching boxes will be hidden,
        and new view will be a white list view.'''
        patchbay_views.new_view(
            self, view_number=view_number, exclusive_with=exclusive_with)

    def rename_current_view(self, new_name: str):
        patchbay_views.rename_current_view(self, new_name)

    def change_view(self, view_number: int):
        patchbay_views.change_view(self, view_number)

    def remove_view(self, view_number: int):
        patchbay_views.remove_view(self, view_number)

    def clear_absents_in_view(self, only_current_ptv=False):
        patchbay_views.clear_absents_in_view(
            self, only_current_ptv=only_current_ptv)

    def remove_all_other_views(self):
        patchbay_views.remove_all_other_views(self)

    def change_view_number(self, new_num: int):
        patchbay_views.change_view_number(self, new_num)

    def write_view_data(
            self, view_number: int, name: str | None =None,
            port_types: PortTypesViewFlag | None =None,
            white_list_view=False):
        patchbay_views.write_view_data(
            self, view_number, name=name, port_types=port_types,
            white_list_view=white_list_view)

    def arrange_follow_signal(self):
        with CancellableAction(self, CancelOp.VIEW) as a:
            a.name = _translate(
                'arrange', 'Arrange: follow the signal chain')
            patchcanvas.arrange_follow_signal()

            # boxes will be at a completely different place after arrange
            # it takes no sense to keep positions of absent boxes
            self.clear_absents_in_view(only_current_ptv=True)

    def arrange_face_to_face(self):
        with CancellableAction(self, CancelOp.VIEW) as a:
            a.name = _translate(
                'arrange', 'Arrange: Two columns facing each other')
            patchcanvas.arrange_face_to_face()

            # boxes will be at a completely different place after arrange
            # it takes no sense to keep positions of absent boxes
            self.clear_absents_in_view(only_current_ptv=True)

    def set_a2j_grouped(self, yesno: int):
        if self.group_a2j_hw != bool(yesno):
            self.group_a2j_hw = bool(yesno)
            self.refresh()

    def set_alsa_midi_enabled(self, yesno: int):
        if self.alsa_midi_enabled != bool(yesno):
            self.alsa_midi_enabled = bool(yesno)
            self.change_port_types_view(PortTypesViewFlag.ALL, force=True)
            self.cancel_mng.reset()

    def set_group_shadows(self, yesno: int):
        patchcanvas.options.show_shadows = bool(yesno)
        self.remove_and_add_all()

    def set_auto_select_items(self, yesno: int):
        patchcanvas.set_auto_select_items(bool(yesno))

    def change_naming(self, naming: Naming):
        if naming is self.naming:
            return

        groups_dict = dict[Group, str]()
        ports_dict = dict[Port, str]()

        for group in self.groups:
            groups_dict[group] = group.cnv_name
            for port in group.ports:
                ports_dict[port] = port.cnv_name

        self.naming = naming

        with CanvasOptimizeIt(self, auto_redraw=True):
            for group in self.groups:
                if not group.in_canvas:
                    continue

                if group.cnv_name != groups_dict[group]:
                    group.rename_in_canvas()

                for port in group.ports:
                    if not port.in_canvas:
                        continue
                    if port.cnv_name != ports_dict[port]:
                        port.rename_in_canvas()

    def change_jack_export_naming(self, naming: Naming):
        self.jack_export_naming = naming

    def export_custom_names_to_jack(self):
        pass

    def import_pretty_names_from_jack(self):
        pass

    def change_theme(self, theme_name: str):
        if not theme_name:
            return

        for connection in self.connections:
            connection.in_canvas = False

        for group in self.groups:
            for portgroup in group.portgroups:
                portgroup.in_canvas = False
            for port in group.ports:
                port.in_canvas = False
            group.in_canvas = False

        patchcanvas.clear_all()
        patchcanvas.change_theme(theme_name)

        with CanvasOptimizeIt(self):
            for group in self.groups:
                group.add_to_canvas()
                for port in group.ports:
                    port.add_to_canvas()
                for portgroup in group.portgroups:
                    portgroup.add_to_canvas()

            for connection in self.connections:
                connection.add_to_canvas()

        patchcanvas.redraw_all_groups()

    def set_elastic_canvas(self, yesno: int):
        patchcanvas.set_elastic(bool(yesno))

    def set_borders_navigation(self, yesno: int):
        patchcanvas.set_borders_navigation(bool(yesno))

    def set_prevent_overlap(self, yesno: int):
        patchcanvas.set_prevent_overlap(bool(yesno))

    def set_zoom(self, zoom: float):
        patchcanvas.canvas.scene.set_zoom_ratio(zoom)

    def zoom_reset(self):
        patchcanvas.zoom_reset()

    def zoom_fit(self):
        patchcanvas.zoom_fit()

    def refresh(self):
        self.clear_all()

    def set_group_uuid_from_name(self, client_name: str, uuid: int):
        patchbay_batches.set_group_uuid_from_name(self, client_name, uuid)

    def add_port(self, name: str, port_type: PortType, flags: int, uuid: int):
        'add port to the global patch, it will be applied later by batches'
        patchbay_batches.add_port(self, name, port_type, flags, uuid)

    def remove_port(self, name: str):
        'remove port from the global patch, will be applied later by batches'
        patchbay_batches.remove_port(self, name)

    def rename_port(self, name: str, new_name: str, uuid=0):
        'rename port in the global patch, will be applied later by batches'
        patchbay_batches.rename_port(self, name, new_name, uuid=uuid)

    def metadata_update(
            self, uuid: int, key: str, value: str):
        '''remember metadata, will be applied later by batches'''
        return patchbay_batches.metadata_update(self, uuid, key, value)

    def add_connection(self, port_out_name: str, port_in_name: str):
        '''add connection to the global patch,
        will be applied later by batches'''
        patchbay_batches.add_connection(self, port_out_name, port_in_name)

    def remove_connection(self, port_out_name: str, port_in_name: str):
        '''remove connection from the global patch,
        will be applied later by batches'''
        patchbay_batches.remove_connection(self, port_out_name, port_in_name)

    def disannounce(self):
        self.clear_all()

    @in_main_thread()
    def server_started(self):
        self.server_is_started = True
        if self._tools_widget is not None:
            self._tools_widget.set_jack_running(True)

        self.clear_all()
        # if this function is executed, all graph will appear just after

    @in_main_thread()
    def server_stopped(self):
        self.server_is_started = False
        if self._tools_widget is not None:
            self._tools_widget.set_jack_running(
                False, use_alsa_midi=self.alsa_midi_enabled)

        self.clear_all()
        if self.alsa_midi_enabled:
            self.refresh()

    @in_main_thread()
    def server_lose(self):
        self.server_is_started = False

        if self._tools_widget is not None:
            self._tools_widget.set_jack_running(False)

        self.clear_all()
        if self.alsa_midi_enabled:
            self.refresh()

        if self.main_win is not None:
            ret = QMessageBox.critical(
                self.main_win,
                _translate('patchbay', "JACK server lose"),
                _translate('patchbay', "JACK server seems to be "
                           "totally busy... ;("))

    @in_main_thread()
    def set_dsp_load(self, dsp_load: int):
        if self._tools_widget is not None:
            self._tools_widget.set_dsp_load(dsp_load)

    @in_main_thread()
    def add_xrun(self):
        if self._tools_widget is not None:
            self._tools_widget.add_xrun()

    @in_main_thread()
    def refresh_transport(self, transport_position: TransportPosition):
        if self._tools_widget is not None:
            self._tools_widget.refresh_transport(transport_position)

    def change_buffersize(self, buffer_size: int):
        '''Inherit this'''

    def transport_play_pause(self, play: bool):
        '''Inherit this'''

    def transport_stop(self):
        '''Inherit this'''

    def transport_relocate(self, frame: int):
        '''Inherit this'''

    def change_tools_displayed(self, tools_displayed: ToolDisplayed):
        if self._tools_widget is not None:
            self._tools_widget.change_tools_displayed(tools_displayed)

        self._tools_displayed = tools_displayed

    def redraw_all_groups(self):
        patchcanvas.redraw_all_groups()

    def filter_groups(self, text: str, n_select=0) -> int:
        '''Semi hides groups not matching with text
        and returns number of matching boxes.'''
        opac_grp_ids = set()

        for group in self.groups:
            if (text.lower() not in group.name.lower()
                    and text.lower() not in group.graceful_name.lower()):
                opac_grp_ids.add(group.group_id)

        patchcanvas.semi_hide_groups(opac_grp_ids)

        n_boxes = 0

        for group in self.groups:
            if group.group_id not in opac_grp_ids:
                n_grp_boxes = group.get_number_of_boxes()

                if n_select > n_boxes and n_select <= n_boxes + n_grp_boxes:
                    group.select_filtered_box(n_select - n_boxes)
                n_boxes += n_grp_boxes

        return n_boxes

    def set_semi_hide_opacity(self, opacity: float):
        patchcanvas.set_semi_hide_opacity(opacity)

    def set_aliasing_reason(self, aliasing_reason: AliasingReason, yesno: bool):
        patchcanvas.set_aliasing_reason(aliasing_reason, yesno)

    def start_aliasing_check(self, aliasing_reason: AliasingReason):
        patchcanvas.start_aliasing_check(aliasing_reason)

    @in_main_thread()
    def buffer_size_changed(self, buffer_size: int):
        if self._tools_widget is not None:
            self._tools_widget.set_buffer_size(buffer_size)

    @in_main_thread()
    def sample_rate_changed(self, samplerate: int):
        if self._tools_widget is not None:
            self._tools_widget.set_samplerate(samplerate)

    def _delayed_orders_timeout(self):
        '''This method is called by the QTimer self._delayed_orders_timer
        when no graph event happens during 50ms.
        It executes in the main thread all methods called since last time,
        then, it updates the canvas with new contents.'''
        patchbay_batches.delayed_orders_timeout(self)

    def apply_delayed_changes_now(self):
        self._delayed_orders_timer.stop()
        self._delayed_orders_timeout()

    def export_to_patchichi_json(self, path: Path, editor_text='') -> bool:
        return export_to_patchichi_json(self, path, editor_text)

    def key_press_event(self, event: QKeyEvent):
        if not event.modifiers() & Qt.KeyboardModifier.ControlModifier:
            if event.modifiers() & Qt.KeyboardModifier.AltModifier:
                if Qt.Key.Key_0 <= event.key() <= Qt.Key.Key_9:
                    new_num = event.key() - Qt.Key.Key_0
                    if self.views.get(new_num) is None:
                            cancel_op = CancelOp.ALL_VIEWS
                    else:
                        cancel_op = CancelOp.VIEW_CHOICE

                    with CancellableAction(self, cancel_op) as a:
                        a.name = _translate('patchbay', 'Change view %i -> %i') \
                            % (self.view_number, new_num)
                        self.change_view(new_num)
            return

        if event.modifiers() & Qt.KeyboardModifier.AltModifier:
            if event.key() == Qt.Key.Key_A:
                self.arrange_follow_signal()
            elif event.key() == Qt.Key.Key_Q:
                self.arrange_face_to_face()

        else:
            if event.key() == Qt.Key.Key_Z:
                if event.modifiers() & Qt.KeyboardModifier.ShiftModifier:
                    self.cancel_mng.redo()
                else:
                    self.cancel_mng.undo()



from dataclasses import dataclass
import logging
from typing import TYPE_CHECKING

from qtpy.QtCore import QRectF, QPointF, QMarginsF, Qt

from patshared import PortMode

from ..box_widget import BoxWidget
from ..init_values import canvas, options, Direction, BoxHidding
from .. import grid

from .scene_utils import MovingBox

if TYPE_CHECKING:
    from . import PatchScene


_logger = logging.getLogger(__name__)


@dataclass
class ToMoveBox:
    directions: list[Direction]
    box: BoxWidget
    rect: QRectF
    pusher_box: BoxWidget
    pusher_rect: QRectF

    def __lt__(self, other: 'ToMoveBox'):
        if self.directions != other.directions:
            return self.directions < other.directions

        # should not happen
        if not self.directions:
            return True

        last_direc = self.directions[-1]
        match last_direc:
            case Direction.LEFT:
                return self.rect.right() > other.rect.right()
            case Direction.UP:
                return self.rect.bottom() > other.rect.bottom()
            case Direction.RIGHT:
                return self.rect.left() < other.rect.left()
            case Direction.DOWN:
                return self.rect.top() < other.rect.top()
            case _:
                # should not happen
                return True


def _get_direction(fixed_rect: QRectF, moving_rect: QRectF,
                   parent_directions: list[Direction] =[]) -> Direction:
    if (moving_rect.top() <= fixed_rect.center().y() <= moving_rect.bottom()
            or (fixed_rect.top()
                <= moving_rect.center().y()
                <= fixed_rect.bottom())):
        if (fixed_rect.right() < moving_rect.center().x()
                and fixed_rect.center().x() < moving_rect.left()):
            if Direction.LEFT in parent_directions:
                return Direction.LEFT
            return Direction.RIGHT

        if (fixed_rect.left() > moving_rect.center().x()
                and fixed_rect.center().x() > moving_rect.right()):
            if Direction.RIGHT in parent_directions:
                return Direction.RIGHT
            return Direction.LEFT

    if fixed_rect.center().y() <= moving_rect.center().y():
        if Direction.UP in parent_directions:
            return Direction.UP
        return Direction.DOWN

    if Direction.DOWN in parent_directions:
        return Direction.DOWN
    return Direction.UP

def _repulse(
        direction: Direction, pusher: BoxWidget | QRectF,
        pushed: BoxWidget | QRectF, pusher_port_mode: PortMode,
        pushed_port_mode: PortMode) -> QRectF:
    '''returns a QRectF to be placed at side of fixed_rect
    where fixed_rect is an already determinated futur place
    for a box'''
    if isinstance(pusher, BoxWidget):
        pusher_rect = pusher.boundingRect().translated(pusher.pos())
    else:
        pusher_rect = pusher

    if isinstance(pushed, BoxWidget):
        rect = pushed.boundingRect().translated(pushed.pos())
    else:
        rect = pushed

    x = rect.left()
    y = rect.top()

    spacing = canvas.theme.box_spacing
    spacing_hor = canvas.theme.box_spacing_horizontal

    match direction:
        case Direction.LEFT | Direction.RIGHT:
            if direction is Direction.LEFT:
                if (pusher_port_mode & PortMode.INPUT
                        or pushed_port_mode & PortMode.OUTPUT):
                    x = grid.previous_left(
                        pusher_rect.left() - rect.width() - spacing_hor)
                else:
                    x = grid.previous_left(
                        pusher_rect.left() - rect.width() - spacing)
            else:
                if (pusher_port_mode & PortMode.OUTPUT
                        or pushed_port_mode & PortMode.INPUT):
                    x = grid.next_left(pusher_rect.right() + spacing_hor)
                else:
                    x = grid.next_left(pusher_rect.right() + spacing)

        case Direction.UP | Direction.DOWN:
            if direction is Direction.UP:
                y = grid.previous_top(
                    pusher_rect.top() - rect.height() - spacing)
            else:
                y = grid.next_top(pusher_rect.bottom() + spacing)

    return QRectF(float(x), float(y), rect.width(), rect.height())

def _rect_may_have_to_move_from(
        repulser_rect: QRectF, rect: QRectF, margins: QMarginsF) -> bool:
    return rect.intersects(repulser_rect.marginsAdded(margins))

def _rect_has_to_move_from(
        repulser_rect: QRectF, rect: QRectF,
        repulser_port_mode: PortMode, rect_port_mode: PortMode) -> bool:
    left_spacing = right_spacing = canvas.theme.box_spacing

    if (repulser_port_mode & PortMode.INPUT
            or rect_port_mode & PortMode.OUTPUT):
        left_spacing = canvas.theme.box_spacing_horizontal

    if (repulser_port_mode & PortMode.OUTPUT
            or rect_port_mode & PortMode.INPUT):
        right_spacing = canvas.theme.box_spacing_horizontal

    return rect.intersects(
        repulser_rect.adjusted(
            - left_spacing, - canvas.theme.box_spacing,
            right_spacing, canvas.theme.box_spacing))

def _get_to_move_boxes_from_repulse_boxes(
        scene: 'PatchScene', pusher_boxes: list[BoxWidget],
        wanted_direction: Direction) -> \
            tuple[list[ToMoveBox], dict[BoxWidget, QRectF]]:
    box_spacing = canvas.theme.box_spacing
    box_spacing_hor = canvas.theme.box_spacing_horizontal
    normal_margins = QMarginsF(
        box_spacing_hor,box_spacing,
        box_spacing_hor, box_spacing)

    to_move_boxes = list[ToMoveBox]()
    pushers = dict[BoxWidget, QRectF]()
    wanted_directions = [wanted_direction]
    
    for pusher_box in pusher_boxes:
        # if box is already moving, consider its end position
        moving_box = scene.move_boxes.get(pusher_box)
        if moving_box is None:
            pusher_rect = pusher_box.after_wrap_rect().translated(
                pusher_box.pos())
        else:
            pusher_rect = moving_box.final_rect
            if pusher_rect.isNull():
                # if this box is joining or hidding,
                # it will be removed soon
                # so, it has not to be a repulser.
                continue

        pushers[pusher_box] = pusher_rect
        pusheds = dict[BoxWidget, QRectF]()

        search_rect = pusher_rect.marginsAdded(normal_margins)

        # search intersections in non moving boxes
        for candidate_box in scene.items(
                search_rect, Qt.ItemSelectionMode.IntersectsItemShape,
                Qt.SortOrder.AscendingOrder):
            if not isinstance(candidate_box, BoxWidget):
                continue

            if (candidate_box in pusher_boxes
                    or candidate_box in [b.box for b in to_move_boxes]
                    or candidate_box in scene.move_boxes):
                continue

            pushed_rect = candidate_box.sceneBoundingRect()

            if _rect_has_to_move_from(
                    pusher_rect, pushed_rect,
                    pusher_box.current_port_mode,
                    candidate_box.current_port_mode):
                pusheds[candidate_box] = pushed_rect

        # search intersections in moving boxes
        for candidate_box, moving_box in scene.move_boxes.items():
            if (candidate_box in pusher_boxes
                    or moving_box.final_rect.isNull()
                    or candidate_box in [b.box for b in to_move_boxes]):
                continue

            pushed_rect = moving_box.final_rect

            if _rect_has_to_move_from(
                    pusher_rect, pushed_rect,
                    pusher_box.current_port_mode,
                    candidate_box.current_port_mode):
                pusheds[candidate_box] = pushed_rect
                
        for pushed_box, pushed_rect in pusheds.items():
            # evaluate in which direction should go the box
            direction = _get_direction(
                pusher_rect, pushed_rect, wanted_directions)
            to_move_boxes.append(
                ToMoveBox([direction], pushed_box, pushed_rect,
                          pusher_box, pusher_rect))

    to_move_boxes.sort()
    return to_move_boxes, pushers

def _get_to_move_boxes_in_full_repulse(
        scene: 'PatchScene', mov_repulsables: list[MovingBox]) -> \
            tuple[list[ToMoveBox], dict[BoxWidget, QRectF]]:
    box_spacing = canvas.theme.box_spacing
    box_spacing_hor = canvas.theme.box_spacing_horizontal
    normal_margins = QMarginsF(
        box_spacing_hor, box_spacing,
        box_spacing_hor, box_spacing)

    # in full repulse, the moving box for this box
    # is the first of the list (optimisation).
    pusher_box = mov_repulsables[0].widget
    pusher_rect = mov_repulsables[0].final_rect
    scene._full_repulse_boxes.add(pusher_box)

    pusheds = dict[BoxWidget, QRectF]()
    to_move_boxes = list[ToMoveBox]()

    for moving_box in mov_repulsables[1:]:
        if moving_box.widget in [b.box for b in to_move_boxes]:
            continue

        if not moving_box.final_rect.intersects(
                pusher_rect.marginsAdded(normal_margins)):
            continue

        if _rect_has_to_move_from(
                pusher_rect, moving_box.final_rect,
                pusher_box.current_port_mode,
                moving_box.widget.current_port_mode):
            pusheds[moving_box.widget] = moving_box.final_rect

    for pushed_box, pushed_rect in pusheds.items():
        # evaluate in which direction should go the box
        direction = _get_direction(pusher_rect, pushed_rect)
        to_move_boxes.append(
            ToMoveBox([direction], pushed_box, pushed_rect,
                      pusher_box, pusher_rect))

    to_move_boxes.sort()
    return to_move_boxes, {pusher_box: pusher_rect}

def deplace_boxes_from_repulsers(
        scene: 'PatchScene', repulser_boxes: list[BoxWidget],
        wanted_direction=Direction.NONE,
        mov_repulsables: list[MovingBox] | None =None):
    '''Change the place of boxes in order to have no box overlapping
    other boxes.'''
    if not options.prevent_overlap:
        return

    box_spacing = canvas.theme.box_spacing
    box_spacing_hor = canvas.theme.box_spacing_horizontal

    if mov_repulsables is not None:
        to_move_boxes, pushers = _get_to_move_boxes_in_full_repulse(
            scene, mov_repulsables)
    else:
        _logger.debug(f'move boxes from {repulser_boxes}')
        to_move_boxes, pushers = _get_to_move_boxes_from_repulse_boxes(
            scene, repulser_boxes, wanted_direction)

    normal_margins = QMarginsF(
        box_spacing_hor, box_spacing,
        box_spacing_hor, box_spacing)

    # !!! to_move_boxes list is dynamic
    # elements can be added to the list while iteration !!!
    for to_move_box in to_move_boxes:
        box, rect, pusher_box, pusher_rect = \
            to_move_box.box, to_move_box.rect, \
            to_move_box.pusher_box, to_move_box.pusher_rect

        directions = to_move_box.directions.copy()
        new_direction = _get_direction(pusher_rect, rect, directions)
        directions.append(new_direction)

        # calculate the new position of the box repulsed by its repulser
        pushed_rect = _repulse(
            new_direction, pusher_rect, rect,
            pusher_box.current_port_mode, box.current_port_mode)

        active_pusher_boxes = set[BoxWidget]()

        # while there is a pusher at new box position
        # move the future box position
        while True:
            # list just here to prevent infinite loop
            # we save the repulsers that already have moved the rect
            for pusher_box, pusher_rect in pushers.items():
                if not pushed_rect.intersects(
                        pusher_rect.marginsAdded(normal_margins)):
                    continue
                
                if _rect_has_to_move_from(
                        pusher_rect, pushed_rect,
                        pusher_box.current_port_mode,
                        box.current_port_mode):

                    if pusher_box in active_pusher_boxes:
                        continue
                    active_pusher_boxes.add(pusher_box)

                    # new_direction = _get_direction(
                    #     pusher_rect, pushed_rect, directions)
                    pushed_rect = _repulse(
                        directions[-1], pusher_rect, pushed_rect,
                        # new_direction, pusher_rect, pushed_rect,
                        pusher_box.current_port_mode,
                        box.current_port_mode)
                    directions.append(new_direction)
                    break
            else:
                break

        # Now we know where the box will be definitely positioned
        # So, this is now a pusher for other boxes
        pushers[box] = pushed_rect
        scene._full_repulse_boxes.add(box)

        # check which existing boxes exists at the new place of the box
        # and add them to this to_move_boxes iteration
        adding_list = list[ToMoveBox]()

        if mov_repulsables is not None:
            for moving_box in mov_repulsables:
                mv_box = moving_box.widget

                if (mv_box in pushers
                        or mv_box in [b.box for b in to_move_boxes]):
                    continue

                if not pushed_rect.intersects(
                        moving_box.final_rect.marginsAdded(normal_margins)):
                    continue

                if _rect_has_to_move_from(
                        pushed_rect, moving_box.final_rect,
                        to_move_box.box.current_port_mode,
                        mv_box.current_port_mode):
                    adding_list.append(
                        ToMoveBox(directions, moving_box.widget,
                                  moving_box.final_rect, box, pushed_rect))
        else:
            search_rect = pushed_rect.marginsAdded(normal_margins)
            for candidate_box in scene.items(search_rect):
                if not isinstance(candidate_box, BoxWidget):
                    continue
                if (candidate_box in pushers
                        or candidate_box in [b.box for b in to_move_boxes]
                        or candidate_box in scene.move_boxes):
                    continue

                candidate_rect = candidate_box.sceneBoundingRect()
                if _rect_has_to_move_from(
                        pushed_rect, candidate_rect,
                        to_move_box.box.current_port_mode,
                        candidate_box.current_port_mode):
                    adding_list.append(
                        ToMoveBox(directions, candidate_box, candidate_rect,
                                  box, pushed_rect))

            for mv_box, moving_box in scene.move_boxes.items():
                if (mv_box in pushers
                        or moving_box.final_rect.isNull()
                        or mv_box in [b.box for b in to_move_boxes]):
                    continue

                if _rect_has_to_move_from(
                        pushed_rect, moving_box.final_rect,
                        to_move_box.box.current_port_mode,
                        mv_box.current_port_mode):
                    adding_list.append(
                        ToMoveBox(directions, moving_box.widget,
                                  moving_box.final_rect, box, pushed_rect))

        adding_list.sort()

        for to_move_box in adding_list:
            to_move_boxes.append(to_move_box)

        # now we decide where the box is moved
        scene.add_box_to_animation(
            box, int(pushed_rect.left()), int(pushed_rect.top()))

def full_repulse(scene: 'PatchScene'):
    if not options.prevent_overlap:
        return

    scene._full_repulse_boxes.clear()

    # add all boxes to animation (this optimize the repulse algorythm)
    for box in canvas.list_boxes():
        if box.isVisible() and box not in scene.move_boxes:
            scene.add_box_to_animation(box, *box.top_left())

    # Now, all boxes are in self.move_boxes
    moving_boxes = [mb for b, mb in scene.move_boxes.items()
                    if (not mb.final_rect.isNull()
                        and b.isVisible())]

    while moving_boxes:
        deplace_boxes_from_repulsers(
            scene, [moving_boxes[0].widget], mov_repulsables=moving_boxes)
        to_rm_movboxes = [mb for mb in moving_boxes
                          if mb.widget in scene._full_repulse_boxes]
        for to_rm_mb in to_rm_movboxes:
            moving_boxes.remove(to_rm_mb)

    scene._full_repulse_boxes.clear()

def bring_neighbors_and_deplace_boxes(
        scene: 'PatchScene', box_widget: BoxWidget, ex_rect: QRectF):
    if not options.prevent_overlap:
        return

    neighbors = [box_widget]
    limit_top = ex_rect.top()
    less_y = ex_rect.height() - box_widget.after_wrap_rect().height()

    box_spacing = canvas.theme.box_spacing

    for neighbor in neighbors:
        if neighbor is box_widget:
            srect = ex_rect
        else:
            moving_box = scene.move_boxes.get(neighbor)
            if moving_box is None:
                srect = neighbor.after_wrap_rect().translated(
                    neighbor.pos())
            else:
                srect = moving_box.final_rect

        for item in scene.items(
                srect.adjusted(0, 0, 0,
                                box_spacing + 1)):
            if item not in neighbors and isinstance(item, BoxWidget):
                nrect = item.after_wrap_rect().translated(item.pos())
                if nrect.top() >= limit_top:
                    neighbors.append(item)

    neighbors.remove(box_widget)

    repulser_boxes = list[BoxWidget]()

    for neighbor in neighbors:
        x, y = neighbor.top_left()
        scene.add_box_to_animation(neighbor, x, int(y - less_y))
        repulser_boxes.append(neighbor)
    repulser_boxes.append(box_widget)

    scene.deplace_boxes_from_repulsers(
        repulser_boxes, wanted_direction=Direction.UP)


'''Contains functions to get the next, previous or nearest point
on the current grid, mainly to know where to place the boxes.'''

from typing import TYPE_CHECKING

from qtpy.QtCore import QPointF

from .init_values import canvas, options

if TYPE_CHECKING:
    from .box_widget import BoxWidget


def nearest(xy: tuple[int, int]) -> tuple[int, int]:
    canvas.ensure_init()
    x, y = xy
    cell_x = options.cell_width
    cell_y = options.cell_height
    margin = canvas.theme.box_spacing // 2

    ret_x = cell_x * (x // cell_x) + margin
    if x - ret_x > cell_x / 2:
        ret_x += cell_x

    ret_y = cell_y * (y // cell_y) + margin
    if y - ret_y > cell_y / 2:
        ret_y += cell_y

    return (ret_x, ret_y)

def nearest_check_others(
        xy: tuple[int, int], orig_box: 'BoxWidget') -> tuple[int, int]:
    '''return the pos for a just moved box,
    may be not exactly the nearest point on grid,
    to prevent unwanted other boxes move.'''
    canvas.ensure_init()
    spacing = canvas.theme.box_spacing
    check_rect = orig_box.boundingRect().translated(QPointF(*xy))
    search_rect = check_rect.adjusted(- spacing, - spacing, spacing, spacing)

    boxes = [b for b in canvas.scene.list_boxes_at(search_rect)
             if b is not orig_box]
    x, y = xy
    new_x, new_y = nearest(xy)

    for box in boxes:
        rect = box.sceneBoundingRect()

        if (previous_top(y)
                == previous_top(rect.bottom())):
            return (new_x, previous_top(y) + options.cell_height)

        if (next_bottom(check_rect.bottom())
                == next_bottom(rect.top())):
            return (new_x, next_top(y) - options.cell_height)

    return nearest(xy)

def previous_left(x: int | float) -> int:
    canvas.ensure_init()
    cell_x = options.cell_width
    margin = canvas.theme.box_spacing / 2

    ret = int(cell_x * (x // cell_x) + margin)
    if ret > x:
        ret -= cell_x

    return ret

def next_left(x: int | float) -> int:
    canvas.ensure_init()
    cell_x = options.cell_width
    margin = canvas.theme.box_spacing / 2

    ret = int(cell_x * (x // cell_x) + margin)
    if ret < x:
        ret += cell_x

    return ret

def previous_top(y: int | float) -> int:
    canvas.ensure_init()
    cell_y = options.cell_height
    margin = canvas.theme.box_spacing / 2

    ret = int(cell_y * (y // cell_y) + margin)
    if ret > y:
        ret -= cell_y

    return ret

def next_top(y: int | float) -> int:
    canvas.ensure_init()
    cell_y = options.cell_height
    margin = canvas.theme.box_spacing / 2

    ret = int(cell_y * ((y - 1) // cell_y) + margin)
    if ret < y:
        ret += cell_y

    return ret

def next_bottom(y: int | float) -> int:
    canvas.ensure_init()
    cell_y = options.cell_height
    margin = canvas.theme.box_spacing / 2

    ret = int(cell_y * (1 + y // cell_y) - margin)
    if ret < y:
        ret += cell_y

    return ret

def next_width(width: int | float) -> int:
    canvas.ensure_init()
    cell_x = options.cell_width
    box_spacing = canvas.theme.box_spacing
    ret = cell_x * (1 + (width // cell_x)) - box_spacing
    while ret < width:
        ret += cell_x

    return int(ret)

def next_height(height: int | float) -> int:
    canvas.ensure_init()
    cell_y = options.cell_height
    box_spacing = canvas.theme.box_spacing
    ret = cell_y * (1 + (height // cell_y)) - box_spacing
    while ret < height:
        ret += cell_y

    return int(ret)



from .style_attributer import StyleAttributer
from .style_attributers import BoxStyleAttributer, UslStyleAttributer
from .theme_structs import Margin, BorderMode, Align
from .theme_utils import to_qcolor
from .theme import Theme



from qtpy.QtCore import QPointF, QRectF

from ..box_widget import BoxWidget
from ..init_values import BoxHidding


class MovingBox:
    widget: BoxWidget
    from_pt: QPointF
    to_pt: QPointF
    final_rect: QRectF
    is_joining: bool
    is_wrapping: bool
    hidding_state: BoxHidding
    needs_move: bool

    def __init__(self, widget: BoxWidget):
        self.widget = widget
        self.from_pt = QPointF(*widget.top_left())
        self.to_pt = QPointF(*widget.top_left())
        self.final_rect = widget.after_wrap_rect().translated(self.to_pt)
        self.is_joining = False
        self.is_wrapping = False
        self.hidding_state = BoxHidding.NONE
        self.needs_move = False

    @property
    def is_useful(self) -> bool:
        if self.needs_move or self.is_wrapping:
            return True
        return self.hidding_state is not BoxHidding.NONE
    
    @property
    def max_distance(self) -> float:
        return max(abs(self.to_pt.x() - self.from_pt.x()),
                   abs(self.to_pt.y() - self.from_pt.y()))



from qtpy.QtGui import QColor

class Color:
    def __init__(self, name: str):
        self.name = name
        self.exists_in = set[str]()
        self.equivalents = set[tuple[str, float]]()


def compare_colors(colors: dict[str, Color]):
    cols = [QColor(color_name) for color_name in colors.keys()]
    cols = sorted(cols, key=QColor.lightness)
    
    for i, col in enumerate(cols):
        if col.lightness() == 0:
            continue

        for col_ in cols[i+1:]:
            if col_.lightnessF() == 1.0:
                continue
            
            ratio = col_.lightnessF() / col.lightnessF()
            rounded_ratio = round(100 * ratio)
            reds, greens, blues = set[int](), set[int](), set[int]()
            for epsilon in (-1, 0, 1):
                light_col = col.lighter(rounded_ratio + epsilon)
                reds.add(light_col.red())
                greens.add(light_col.green())
                blues.add(light_col.blue())
                
            if (col_.red() in reds
                    and col_.green() in greens
                    and col_.blue() in blues):
                colors[col_.name()].equivalents.add((col.name(), ratio))
                colors[col.name()].equivalents.add((col_.name(), 1 / ratio))
        


import logging
import pickle
from typing import TypeAlias

from .. import xdg


_logger = logging.getLogger(__name__)

TitleCache: TypeAlias = dict[str, dict[int, list[dict[str, int]]]]

# if for some reason cache may be incompatible with this version
# of the patchbay, we need to discard the cache files.
CACHE_VERSION = (1, 4)

title_templates_cache: dict[str, dict[str, dict[str, TitleCache]]] = \
    {'CACHE_VERSION': CACHE_VERSION} # type:ignore
font_metrics_cache: dict[str, dict[str, dict[str, dict[str, float]]]] = \
    {'CACHE_VERSION': CACHE_VERSION} # type:ignore


def load():
    cache_file = xdg.xdg_cache_home() / 'HoustonPatchbay' / 'patchbay_titles'
    if not cache_file.is_file():
        return

    with open(cache_file, 'rb') as f:
        try:
            title_templates_cache_ = pickle.load(f)
            assert title_templates_cache_['CACHE_VERSION'] == CACHE_VERSION
            title_templates_cache.clear()
            title_templates_cache.update(title_templates_cache_)
        except:
            _logger.warning(f"failed to load cache {cache_file}")
            return

    font_cache_file = xdg.xdg_cache_home() / 'HoustonPatchbay' / 'patchbay_fonts'
    if not font_cache_file.is_file():
        return

    with open(font_cache_file, 'rb') as f:
        try:
            font_metrics_cache_ = pickle.load(f)
            assert font_metrics_cache_['CACHE_VERSION'] == CACHE_VERSION
            font_metrics_cache.clear()
            font_metrics_cache.update(font_metrics_cache_)
        except:
            _logger.error(f"failed to load font cache {font_cache_file}")
            return

def save():
    cache_dir = xdg.xdg_cache_home() / 'HoustonPatchbay'
    if not cache_dir.is_dir():
        try:
            cache_dir.mkdir(parents=True, exist_ok=True)
        except:
            return

    with open(cache_dir / 'patchbay_titles', 'wb') as f:
        pickle.dump(title_templates_cache, f)

    with open(cache_dir / 'patchbay_fonts', 'wb') as f:
        pickle.dump(font_metrics_cache, f)
        
def get_font_metrics_cache(
        font_name: str, font_size: str, font_width: str) -> dict[str, float]:
    if not font_name in font_metrics_cache.keys():
        font_metrics_cache[font_name] = \
            dict[str, dict[str, dict[str, float]]]()

    if not font_size in font_metrics_cache[font_name].keys():
        font_metrics_cache[font_name][font_size] = \
            dict[str, dict[str, float]]()

    if not font_width in font_metrics_cache[font_name][font_size].keys():
        font_metrics_cache[font_name][font_size][font_width] = \
            dict[str, float]()

    return font_metrics_cache[font_name][font_size][font_width]

def get_title_templates_cache(
        font_name: str, font_size: str, font_width: str) -> TitleCache:
    if not font_name in title_templates_cache.keys():
        title_templates_cache[font_name] = \
            dict[str, dict[str, TitleCache]]()

    if not font_size in title_templates_cache[font_name].keys():
        title_templates_cache[font_name][font_size] = dict[str, TitleCache]()

    if not font_width in title_templates_cache[font_name][font_size].keys():
        title_templates_cache[font_name][font_size][font_width] = TitleCache()

    return title_templates_cache[font_name][font_size][font_width]



from typing import TYPE_CHECKING

from qtpy.QtCore import QRectF
from qtpy.QtGui import QPainterPath

from patshared import PortMode

from ..patchcanvas import canvas
from ..theme import StyleAttributer

from .box_utils import BoxStyler, PaintElement, WrappingState

if TYPE_CHECKING:
    from .box_widget import BoxWidget


# theses values are needed to prevent some incorrect painter_path
# united or subtracted results
EPSY = 0.001
EPSD = EPSY * 2.0


def _build_main_path(
        box: 'BoxWidget', pos_dict: dict[str, list[list[float]]],
        selected=False) -> QPainterPath:
    input_segments = pos_dict['input_segments']
    output_segments = pos_dict['output_segments']

    painter_path = QPainterPath()
    theme = box.get_theme()
    if selected:
        theme = theme.selected

    border_radius = theme.border_radius
    port_in_offset = abs(theme.port_in_offset)
    port_out_offset = abs(theme.port_out_offset)
    bore_in = bool(theme.port_in_offset_mode == 'bore')
    bore_out = bool(theme.port_out_offset_mode == 'bore')
    line_hinting = theme.border_width / 2

    rect = QRectF(0.0, 0.0, box._width, box._height)
    rect.adjust(line_hinting, line_hinting, -line_hinting, -line_hinting)

    if border_radius == 0.0:
        painter_path.addRect(rect)
    else:
        painter_path.addRoundedRect(rect, border_radius, border_radius)

    if not bore_in:
        port_in_offset = 0.0
    if not bore_out:
        port_out_offset = 0.0

    if box._wrapping_state is WrappingState.NORMAL:
        # substract rects in the box shape in case of 
        # port_offset (even negativ).
        # logic would want to add rects if port_offset is negativ
        # But that also means that we should change the boudingRect,
        # So we won't.
        if port_in_offset != 0.0:
            for in_segment in input_segments:
                moins_path = QPainterPath()
                moins_path.addRect(QRectF(
                    0.0 - EPSY,
                    in_segment[0] - line_hinting - EPSY,
                    port_in_offset + line_hinting + EPSD,
                    in_segment[1] - in_segment[0] + line_hinting * 2 + EPSD))
                painter_path = painter_path.subtracted(moins_path)

        if port_out_offset != 0.0:
            for out_segment in output_segments:
                moins_path = QPainterPath()
                moins_path.addRect(QRectF(
                    box._width - line_hinting - port_out_offset - EPSY,
                    out_segment[0] - line_hinting - EPSY,
                    port_out_offset + line_hinting + EPSD,
                    out_segment[1] - out_segment[0] + line_hinting * 2 + EPSD))
                painter_path = painter_path.subtracted(moins_path)

        # No rounded corner if the last port is to close from the corner
        if (input_segments
                and box._height - input_segments[-1][1] <= border_radius):
            left_path = QPainterPath()
            left_path.addRect(QRectF(
                0.0 + line_hinting - EPSY,
                max(box._height - border_radius, input_segments[-1][1])
                    + line_hinting - EPSY,
                border_radius + EPSD,
                min(border_radius, box._height - input_segments[-1][1])
                    - 2 * line_hinting + EPSD))
            painter_path = painter_path.united(left_path)

        if (input_segments
                and input_segments[0][0] <= border_radius):
            top_left_path = QPainterPath()
            top_left_path.addRect(QRectF(
                0.0 + line_hinting - EPSY,
                0.0 + line_hinting - EPSY,
                border_radius + EPSD,
                min(border_radius, input_segments[0][0])
                - 2 * line_hinting + EPSD))
            painter_path = painter_path.united(top_left_path)

        if (output_segments
                and box._height - output_segments[-1][1] <= border_radius):
            right_path = QPainterPath()
            right_path.addRect(QRectF(
                box._width - border_radius - line_hinting - EPSY,
                max(box._height - border_radius, output_segments[-1][1])
                    + line_hinting - EPSY,
                border_radius + EPSD,
                min(border_radius, box._height - output_segments[-1][1])
                    - 2 * line_hinting + EPSD))
            painter_path = painter_path.united(right_path)

        if (output_segments
                and output_segments[0][0] <= border_radius):
            top_right_path = QPainterPath()
            top_right_path.addRect(QRectF(
                box._width - line_hinting + EPSY - border_radius,
                0.0 + line_hinting - EPSY,
                border_radius + EPSD,
                min(border_radius, output_segments[0][0])
                - 2 * line_hinting + EPSD))
            painter_path = painter_path.united(top_right_path)

    if box.is_monitor and border_radius:
        if box.current_port_mode is PortMode.OUTPUT:
            left_path = QPainterPath()
            left_path.addRect(QRectF(
                0.0 + line_hinting - EPSY,
                box._height - border_radius - EPSY,
                border_radius + EPSD, border_radius - line_hinting + EPSD))
            painter_path = painter_path.united(left_path)

            top_left_path = QPainterPath()
            top_left_path.addRect(QRectF(
                0.0 + line_hinting - EPSY, 0.0 + line_hinting - EPSY,
                border_radius + EPSD, border_radius - line_hinting + EPSD))
            painter_path = painter_path.united(top_left_path)

        elif box.current_port_mode is PortMode.INPUT:
            right_path = QPainterPath()
            right_path.addRect(QRectF(
                box._width - line_hinting - EPSY - border_radius,
                box._height - border_radius - EPSY,
                border_radius + EPSD, border_radius - line_hinting + EPSD))
            painter_path = painter_path.united(right_path)

            top_right_path = QPainterPath()
            top_right_path.addRect(QRectF(
                box._width - line_hinting - EPSY - border_radius,
                0.0 + line_hinting - EPSY,
                border_radius + EPSD, border_radius - line_hinting + EPSD))
            painter_path = painter_path.united(top_right_path)
    
    return painter_path

def _build_header_path(
        box: 'BoxWidget', header_theme: StyleAttributer,
        usl_border: float, line_hinting: float) -> QPainterPath:
    mg = header_theme.margin
    header_lh = header_theme.border_width * 0.5
    if header_lh == 0:
        # fix artefacts in case there is no border for the header
        border = 0
    else:
        border = line_hinting * 2.0
    
    if box.has_side_title:
        if box.current_port_mode is PortMode.OUTPUT:
            header_rect = QRectF(
                border + header_lh + mg.free_side - EPSY,
                border + header_lh - EPSY,
                box._header_width - mg.sided_width
                    - 2 * header_lh - border + usl_border + EPSD,
                box._height - 2 * (border + header_lh) + EPSD
            )
        else:
            header_rect = QRectF(
                box._width - border - box._header_width + mg.ports_side
                    + header_lh + border - usl_border - EPSY,
                border + header_lh - EPSY,
                box._header_width - mg.sided_width - 2 * header_lh + EPSD,
                box._height - 2 * (border + header_lh) + EPSD
            )

    else:
        header_rect = QRectF(
            border + header_lh - EPSY,
            border + mg.top + header_lh -EPSY,
            box._width - 2 * (border + header_lh) + EPSD,
            box._header_height - mg.height
                - 2 * header_lh - border + usl_border + EPSD)
    
    header_path = QPainterPath()
    header_path.addRect(header_rect)
    return header_path

def _build_gui_button_path(
        box: 'BoxWidget', header_theme: StyleAttributer,
        usl_border: float) -> QPainterPath:
    gui_theme = canvas.theme.gui_button
    gui_path = QPainterPath()
    gmg = gui_theme.margin
    hmg = header_theme.margin
    mg = hmg + gmg
    gui_lh = gui_theme.border_width / 2
    
    if box.has_side_title:
        if box.current_port_mode is PortMode.OUTPUT:
            gui_rect = QRectF(
                usl_border + mg.free_side + gui_lh,
                usl_border + mg.top_side + gui_lh,
                box._header_width - mg.sided_width - 2 * gui_lh,
                box._header_height - mg.sided_height - 2 * gui_lh)
        else:
            gui_rect = QRectF(
                box._width - box._header_width - usl_border + mg.ports_side + gui_lh,
                mg.top_side + usl_border + gui_lh,
                box._header_width - mg.sided_width - gui_lh * 2,
                box._header_height - mg.sided_height - gui_lh * 2)
    else:
        space_left = (box._width - box._header_width - 2 * usl_border) / 2
        gui_rect = QRectF(
            usl_border + mg.sides + space_left + gui_lh,
            usl_border + mg.top + gui_lh,
            box._header_width - mg.width - gui_lh * 2,
            box._header_height - mg.height - gui_lh * 2)
        
    radius = gui_theme.border_radius
    if radius:
        gui_path.addRoundedRect(gui_rect, radius, radius)
    else:
        gui_path.addRect(gui_rect)
        
    return gui_path

def build_painter_path(
        box: 'BoxWidget', pos_dict: dict[str, list[list[float]]],
        selected=False):
    painter_paths = box._painter_paths.get(selected)
    if painter_paths is None:
        painter_paths = box._painter_paths[selected] = \
            dict[PaintElement, QPainterPath]()
    painter_paths.clear()

    main_path = _build_main_path(box, pos_dict, selected)
    painter_paths[PaintElement.MAIN] = main_path

    theme = box.get_theme()
    usl_border = theme.border_width
    if selected:
        theme = theme.selected

    header_theme = box.get_theme(BoxStyler.HEADER)
    if selected:
        header_theme = header_theme.selected

    drill_gui_button = False

    if box._can_handle_gui:
        painter_paths[PaintElement.GUI_BUTTON] = \
            _build_gui_button_path(box, header_theme, usl_border)
        drill_gui_button = canvas.theme.gui_button.drilled

    if header_theme.visible:
        tmp_header_path = _build_header_path(
            box, header_theme, usl_border, theme.border_width / 2)

        if drill_gui_button:
            tmp_header_path = tmp_header_path.subtracted(
                painter_paths[PaintElement.GUI_BUTTON])

        painter_paths[PaintElement.HEADER] = \
            main_path.intersected(tmp_header_path)   
        
        if header_theme.drilled:
            painter_paths[PaintElement.ANTI_HEADER] = \
                main_path.subtracted(tmp_header_path)
